#!/usr/bin/env python3
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import gi
gi.require_version('Gimp', '3.0')
from gi.repository import Gimp
gi.require_version('Gegl', '0.4')
from gi.repository import Gegl
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gio
import time
import sys

def N_(message): return message
def _(message): return GLib.dgettext(None, message)

def foggify(procedure, run_mode, image, drawables, config, data):
    if run_mode == Gimp.RunMode.INTERACTIVE:
        gi.require_version('GimpUi', '3.0')
        from gi.repository import GimpUi

        GimpUi.init('python-fu-foggify')

        dialog = GimpUi.ProcedureDialog(procedure=procedure, config=config)
        dialog.fill(None)
        if not dialog.run():
            dialog.destroy()
            return procedure.new_return_values(Gimp.PDBStatusType.CANCEL, GLib.Error())
        else:
            dialog.destroy()

    color      = config.get_property('color')
    name       = config.get_property('name')
    turbulence = config.get_property('turbulence')
    opacity    = config.get_property('opacity')

    Gimp.context_push()
    image.undo_group_start()

    if image.get_base_type() is Gimp.ImageBaseType.RGB:
        type = Gimp.ImageType.RGBA_IMAGE
    else:
        type = Gimp.ImageType.GRAYA_IMAGE
    for drawable in drawables:
        fog = Gimp.Layer.new(image, name,
                             drawable.get_width(), drawable.get_height(),
                             type, opacity,
                             Gimp.LayerMode.NORMAL)
        fog.fill(Gimp.FillType.TRANSPARENT)
        image.insert_layer(fog, drawable.get_parent(),
                           image.get_item_position(drawable))

        Gimp.context_set_background(color)
        fog.edit_fill(Gimp.FillType.BACKGROUND)

        # create a layer mask for the new layer
        mask = fog.create_mask(0)
        fog.add_mask(mask)

        # add some clouds to the layer
        _, x, y, width, height = mask.mask_intersect()
        if not Gimp.Selection.is_empty(image):
          x = 0
          y = 0
        plasma_filter = Gimp.DrawableFilter.new(mask, "gegl:plasma", "")
        plasma_filter_config = plasma_filter.get_config()
        plasma_filter_config.set_property("seed", int(time.time()))
        plasma_filter_config.set_property("turbulence", turbulence)
        plasma_filter_config.set_property("x", x)
        plasma_filter_config.set_property("y", y)
        plasma_filter_config.set_property("width", width)
        plasma_filter_config.set_property("height", height)
        mask.merge_filter (plasma_filter)

        # apply the clouds to the layer
        fog.remove_mask(Gimp.MaskApplyMode.APPLY)
        fog.set_visible(True)

    Gimp.displays_flush()

    image.undo_group_end()
    Gimp.context_pop()

    return procedure.new_return_values(Gimp.PDBStatusType.SUCCESS, GLib.Error())

class Foggify (Gimp.PlugIn):
    ## GimpPlugIn virtual methods ##
    def do_set_i18n(self, procname):
        return True, 'gimp30-python', None

    def do_query_procedures(self):
        return [ 'python-fu-foggify' ]

    def do_create_procedure(self, name):
        Gegl.init(None)

        _color = Gegl.Color.new("black")
        _color.set_rgba(0.871, 0.457, 0.061, 1.0)

        procedure = Gimp.ImageProcedure.new(self, name,
                                            Gimp.PDBProcType.PLUGIN,
                                            foggify, None)
        procedure.set_image_types("RGB*, GRAY*");
        procedure.set_sensitivity_mask (Gimp.ProcedureSensitivityMask.DRAWABLE |
                                        Gimp.ProcedureSensitivityMask.DRAWABLES)
        procedure.set_documentation (_("Add a layer of fog"),
                                     _("Adds a layer of fog to the image."),
                                     name)
        procedure.set_menu_label(_("_Fog..."))
        procedure.set_attribution("James Henstridge",
                                  "James Henstridge",
                                  "1999,2007")
        procedure.add_menu_path ("<Image>/Filters/Decor")

        procedure.add_string_argument ("name", _("Layer _name"), _("Layer name"),
                                       _("Clouds"), GObject.ParamFlags.READWRITE)
        procedure.add_color_argument ("color", _("_Fog color"), _("Fog color"),
                                      True, _color, GObject.ParamFlags.READWRITE)
        procedure.add_double_argument ("turbulence", _("_Turbulence"), _("Turbulence"),
                                       0.0, 7.0, 1.0, GObject.ParamFlags.READWRITE)
        procedure.add_double_argument ("opacity", _("O_pacity"), _("Opacity"),
                                       0.0, 100.0, 100.0, GObject.ParamFlags.READWRITE)

        return procedure

Gimp.main(Foggify.__gtype__, sys.argv)



#!/usr/bin/env python3

# GIMP Plug-in for the OpenRaster file format
# http://create.freedesktop.org/wiki/OpenRaster

# Copyright (C) 2009 by Jon Nordby <jononor@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Based on MyPaint source code by Martin Renold
# http://gitorious.org/mypaint/mypaint/blobs/edd84bcc1e091d0d56aa6d26637aa8a925987b6a/lib/document.py

import gi
gi.require_version('Gimp', '3.0')
from gi.repository import Gimp
gi.require_version('Gegl', '0.4')
from gi.repository import Gegl
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gio

import os, sys, tempfile, zipfile
import xml.etree.ElementTree as ET

def N_(message): return message
def _(message): return GLib.dgettext(None, message)

NESTED_STACK_END = object()

layermodes_map = {
    "svg:src-over":     Gimp.LayerMode.NORMAL,
    "svg:multiply":     Gimp.LayerMode.MULTIPLY,
    "svg:screen":       Gimp.LayerMode.SCREEN,
    "svg:overlay":      Gimp.LayerMode.OVERLAY,
    "svg:darken":       Gimp.LayerMode.DARKEN_ONLY,
    "svg:lighten":      Gimp.LayerMode.LIGHTEN_ONLY,
    "svg:color-dodge":  Gimp.LayerMode.DODGE,
    "svg:color-burn":   Gimp.LayerMode.BURN,
    "svg:hard-light":   Gimp.LayerMode.HARDLIGHT,
    "svg:soft-light":   Gimp.LayerMode.SOFTLIGHT,
    "svg:difference":   Gimp.LayerMode.DIFFERENCE,
    "svg:color":        Gimp.LayerMode.HSL_COLOR,
    "svg:luminosity":   Gimp.LayerMode.HSV_VALUE,
    "svg:hue":          Gimp.LayerMode.HSV_HUE,
    "svg:saturation":   Gimp.LayerMode.HSV_SATURATION,
    "svg:plus":         Gimp.LayerMode.ADDITION,
}

# There are less svg blending ops than we have GIMP blend modes.
# We are going to map them as closely as possible.
gimp_layermodes_map = {
    Gimp.LayerMode.NORMAL:                  "svg:src-over",
    Gimp.LayerMode.NORMAL_LEGACY:           "svg:src-over",
    Gimp.LayerMode.MULTIPLY:                "svg:multiply",
    Gimp.LayerMode.MULTIPLY_LEGACY:         "svg:multiply",
    Gimp.LayerMode.SCREEN:                  "svg:screen",
    Gimp.LayerMode.SCREEN_LEGACY:           "svg:screen",
    Gimp.LayerMode.OVERLAY:                 "svg:overlay",
    Gimp.LayerMode.OVERLAY_LEGACY:          "svg:overlay",
    Gimp.LayerMode.DARKEN_ONLY:             "svg:darken",
    Gimp.LayerMode.DARKEN_ONLY_LEGACY:      "svg:darken",
    Gimp.LayerMode.LIGHTEN_ONLY:            "svg:lighten",
    Gimp.LayerMode.LIGHTEN_ONLY_LEGACY:     "svg:lighten",
    Gimp.LayerMode.DODGE:                   "svg:color-dodge",
    Gimp.LayerMode.DODGE_LEGACY:            "svg:color-dodge",
    Gimp.LayerMode.BURN:                    "svg:color-burn",
    Gimp.LayerMode.BURN_LEGACY:             "svg:color-burn",
    Gimp.LayerMode.HARDLIGHT:               "svg:hard-light",
    Gimp.LayerMode.HARDLIGHT_LEGACY:        "svg:hard-light",
    Gimp.LayerMode.SOFTLIGHT:               "svg:soft-light",
    Gimp.LayerMode.SOFTLIGHT_LEGACY:        "svg:soft-light",
    Gimp.LayerMode.DIFFERENCE:              "svg:difference",
    Gimp.LayerMode.DIFFERENCE_LEGACY:       "svg:difference",
    Gimp.LayerMode.HSL_COLOR:               "svg:color",
    Gimp.LayerMode.HSL_COLOR_LEGACY:        "svg:color",
    Gimp.LayerMode.HSV_VALUE:               "svg:luminosity",
    Gimp.LayerMode.HSV_VALUE_LEGACY:        "svg:luminosity",
    Gimp.LayerMode.HSV_HUE:                 "svg:hue",
    Gimp.LayerMode.HSV_HUE_LEGACY:          "svg:hue",
    Gimp.LayerMode.HSV_SATURATION:          "svg:saturation",
    Gimp.LayerMode.HSV_SATURATION_LEGACY:   "svg:saturation",
    Gimp.LayerMode.ADDITION:                "svg:plus",
    Gimp.LayerMode.ADDITION_LEGACY:         "svg:plus",

    # FIXME Determine the closest available layer mode
    #       Alternatively we could add additional modes
    #       e.g. something like "gimp:dissolve", this
    #       is what Krita seems to do too
    Gimp.LayerMode.DISSOLVE:                "svg:src-over",
    Gimp.LayerMode.DIVIDE:                  "svg:src-over",
    Gimp.LayerMode.DIVIDE_LEGACY:           "svg:src-over",
    Gimp.LayerMode.BEHIND:                  "svg:src-over",
    Gimp.LayerMode.BEHIND_LEGACY:           "svg:src-over",
    Gimp.LayerMode.GRAIN_EXTRACT:           "svg:src-over",
    Gimp.LayerMode.GRAIN_EXTRACT_LEGACY:    "svg:src-over",
    Gimp.LayerMode.GRAIN_MERGE:             "svg:src-over",
    Gimp.LayerMode.GRAIN_MERGE_LEGACY:      "svg:src-over",
    Gimp.LayerMode.COLOR_ERASE:             "svg:src-over",
    Gimp.LayerMode.COLOR_ERASE_LEGACY:      "svg:src-over",
    Gimp.LayerMode.LCH_HUE:                 "svg:src-over",
    Gimp.LayerMode.LCH_CHROMA:              "svg:src-over",
    Gimp.LayerMode.LCH_COLOR:               "svg:src-over",
    Gimp.LayerMode.LCH_LIGHTNESS:           "svg:src-over",
    Gimp.LayerMode.SUBTRACT:                "svg:src-over",
    Gimp.LayerMode.SUBTRACT_LEGACY:         "svg:src-over",
    Gimp.LayerMode.VIVID_LIGHT:             "svg:src-over",
    Gimp.LayerMode.PIN_LIGHT:               "svg:src-over",
    Gimp.LayerMode.LINEAR_LIGHT:            "svg:src-over",
    Gimp.LayerMode.HARD_MIX:                "svg:src-over",
    Gimp.LayerMode.EXCLUSION:               "svg:src-over",
    Gimp.LayerMode.LINEAR_BURN:             "svg:src-over",
    Gimp.LayerMode.LUMA_DARKEN_ONLY:        "svg:src-over",
    Gimp.LayerMode.LUMA_LIGHTEN_ONLY:       "svg:src-over",
    Gimp.LayerMode.LUMINANCE:               "svg:src-over",
    Gimp.LayerMode.COLOR_ERASE:             "svg:src-over",
    Gimp.LayerMode.ERASE:                   "svg:src-over",
    Gimp.LayerMode.MERGE:                   "svg:src-over",
    Gimp.LayerMode.SPLIT:                   "svg:src-over",
    Gimp.LayerMode.PASS_THROUGH:            "svg:src-over",
}

def reverse_map(mapping):
    return dict((v,k) for k, v in mapping.items())

def get_image_attributes(orafile):
    xml = orafile.read('stack.xml')
    image = ET.fromstring(xml)
    stack = image.find('stack')
    w = int(image.attrib.get('w', ''))
    h = int(image.attrib.get('h', ''))

    return stack, w, h

def get_layer_attributes(layer):
    a = layer.attrib
    path = a.get('src', '')
    name = a.get('name', '')
    x = int(a.get('x', '0'))
    y = int(a.get('y', '0'))
    opac = float(a.get('opacity', '1.0'))
    visible = a.get('visibility', 'visible') != 'hidden'
    m = a.get('composite-op', 'svg:src-over')
    layer_mode = layermodes_map.get(m, Gimp.LayerMode.NORMAL)
    selected = a.get('selected', 'false') == 'true'
    locked = a.get('edit-locked', 'false') == 'true'

    return path, name, x, y, opac, visible, layer_mode, selected, locked

def get_group_layer_attributes(layer):
    a = layer.attrib
    name = a.get('name', '')
    opac = float(a.get('opacity', '1.0'))
    visible = a.get('visibility', 'visible') != 'hidden'
    m = a.get('composite-op', 'svg:src-over')
    layer_mode = layermodes_map.get(m, Gimp.LayerMode.NORMAL)
    selected = a.get('selected', 'false') == 'true'
    locked = a.get('edit-locked', 'false') == 'true'

    return name, 0, 0, opac, visible, layer_mode, selected, locked

def thumbnail_ora(procedure, file, thumb_size, args, data):
    tempdir = tempfile.mkdtemp('gimp-plugin-file-openraster')
    orafile = zipfile.ZipFile(file.peek_path())
    stack, w, h = get_image_attributes(orafile)

    # create temp file
    tmp = os.path.join(tempdir, 'tmp.png')
    with open(tmp, 'wb') as fid:
        fid.write(orafile.read('Thumbnails/thumbnail.png'))

    thumb_file = Gio.file_new_for_path(tmp)
    pdb_proc   = Gimp.get_pdb().lookup_procedure('file-png-load')
    pdb_config = pdb_proc.create_config()
    pdb_config.set_property('run-mode', Gimp.RunMode.NONINTERACTIVE)
    pdb_config.set_property('file', thumb_file)
    result = pdb_proc.run(pdb_config)
    os.remove(tmp)
    os.rmdir(tempdir)

    if (result.index(0) == Gimp.PDBStatusType.SUCCESS):
        img = result.index(1)
        # TODO: scaling

        return Gimp.ValueArray.new_from_values([
            GObject.Value(Gimp.PDBStatusType, Gimp.PDBStatusType.SUCCESS),
            GObject.Value(Gimp.Image, img),
            GObject.Value(GObject.TYPE_INT, w),
            GObject.Value(GObject.TYPE_INT, h),
            GObject.Value(Gimp.ImageType, Gimp.ImageType.RGB_IMAGE),
            GObject.Value(GObject.TYPE_INT, 1)
        ])
    else:
        return procedure.new_return_values(result.index(0), GLib.Error(result.index(1)))

def export_ora(procedure, run_mode, image, file, options, metadata, config, data):
    def write_file_str(zfile, fname, data):
        # work around a permission bug in the zipfile library:
        # http://bugs.python.org/issue3394
        zi = zipfile.ZipInfo(fname)
        zi.external_attr = int("100644", 8) << 16
        zfile.writestr(zi, data)

    Gimp.progress_init("Exporting openraster image")
    tempdir = tempfile.mkdtemp('gimp-plugin-file-openraster')

    # use .tmpsave extension, so we don't overwrite a valid file if
    # there is an exception
    orafile = zipfile.ZipFile(file.peek_path() + '.tmpsave', 'w', compression=zipfile.ZIP_STORED)

    write_file_str(orafile, 'mimetype', 'image/openraster') # must be the first file written

    # build image attributes
    xml_image = ET.Element('image')
    stack = ET.SubElement(xml_image, 'stack')
    a = xml_image.attrib
    a['w'] = str(image.get_width())
    a['h'] = str(image.get_height())

    def store_layer(image, drawable, path):
        tmp = os.path.join(tempdir, 'tmp.png')
        interlace, compression = 0, 2

        width, height = drawable.get_width(), drawable.get_height()
        tmp_img = Gimp.Image.new(width, height, image.get_base_type())
        if (image.get_base_type() == Gimp.ImageBaseType.INDEXED):
            tmp_img.set_palette(image.get_palette())

        tmp_layer = Gimp.Layer.new_from_drawable (drawable, tmp_img)
        tmp_layer.set_offsets (0, 0)
        tmp_img.insert_layer (tmp_layer, None, 0)

        pdb_proc   = Gimp.get_pdb().lookup_procedure('file-png-export')
        pdb_config = pdb_proc.create_config()
        pdb_config.set_property('run-mode', Gimp.RunMode.NONINTERACTIVE)
        pdb_config.set_property('image', tmp_img)
        pdb_config.set_property('file', Gio.File.new_for_path(tmp))
        pdb_config.set_property('options', None)
        pdb_config.set_property('interlaced', interlace)
        pdb_config.set_property('compression', compression)
        # write all PNG chunks except oFFs(ets)
        pdb_config.set_property('bkgd', True)
        pdb_config.set_property('offs', False)
        pdb_config.set_property('phys', True)
        pdb_config.set_property('time', True)
        pdb_config.set_property('save-transparent', True)
        pdb_proc.run(pdb_config)
        if (os.path.exists(tmp)):
            orafile.write(tmp, path)
            os.remove(tmp)
        else:
            print("Error removing ", tmp)

        tmp_img.delete()

    def add_layer(parent, x, y, opac, gimp_layer, path, visible=True):
        store_layer(image, gimp_layer, path)
        # create layer attributes
        layer = ET.Element('layer')
        parent.append(layer)
        a = layer.attrib
        a['src'] = path
        a['name'] = gimp_layer.get_name()
        a['x'] = str(x)
        a['y'] = str(y)
        a['opacity'] = str(opac)
        a['visibility'] = 'visible' if visible else 'hidden'
        a['composite-op'] = gimp_layermodes_map.get(gimp_layer.get_mode(), 'svg:src-over')

        # ORA Specification extensions
        if gimp_layer in image.get_selected_layers():
            a['selected'] = "true"
        if gimp_layer.get_lock_content():
            a['edit-locked'] = "true"
        return layer

    def add_group_layer(parent, opac, gimp_layer, visible=True):
        # create layer attributes
        group_layer = ET.Element('stack')
        parent.append(group_layer)
        a = group_layer.attrib
        a['name'] = gimp_layer.get_name()
        a['opacity'] = str(opac)
        a['visibility'] = 'visible' if visible else 'hidden'
        a['composite-op'] = gimp_layermodes_map.get(gimp_layer.get_mode(), 'svg:src-over')

        # ORA Specification extensions
        if gimp_layer in image.get_selected_layers():
            a['selected'] = "true"
        if gimp_layer.get_lock_content():
            a['edit-locked'] = "true"
        return group_layer


    def enumerate_layers(layers):
        for layer in layers:
            if not layer.is_group():
                yield layer
            else:
                yield layer
                for sublayer in enumerate_layers(layer.get_children()):
                    yield sublayer
                yield NESTED_STACK_END

    # save layers
    parent_groups = []
    i = 0

    layer_stack = image.get_layers()
    # Number of top level layers for tracking progress
    lay_cnt = len(layer_stack)

    for lay in enumerate_layers(layer_stack):
        prev_lay = i
        if lay is NESTED_STACK_END:
            parent_groups.pop()
            continue
        _, x, y = lay.get_offsets()
        opac = lay.get_opacity () / 100.0 # needs to be between 0.0 and 1.0

        if not parent_groups:
            path_name = 'data/{:03d}.png'.format(i)
            i += 1
        else:
            path_name = 'data/{}-{:03d}.png'.format(
                parent_groups[-1][1], parent_groups[-1][2])
            parent_groups[-1][2] += 1

        parent = stack if not parent_groups else parent_groups[-1][0]

        if lay.is_group():
            group = add_group_layer(parent, opac, lay, lay.get_visible())
            group_path = ("{:03d}".format(i) if not parent_groups else
                parent_groups[-1][1] + "-{:03d}".format(parent_groups[-1][2]))
            parent_groups.append([group, group_path , 0])
        else:
            add_layer(parent, x, y, opac, lay, path_name, lay.get_visible())

        if (i > prev_lay):
            Gimp.progress_update(i/lay_cnt)

    # save mergedimage
    thumb = image.duplicate()
    thumb_layer = thumb.merge_visible_layers (Gimp.MergeType.CLIP_TO_IMAGE)
    store_layer (thumb, thumb_layer, 'mergedimage.png')

    # save thumbnail
    w, h = image.get_width(), image.get_height()
    if max (w, h) > 256:
        # should be at most 256x256, without changing aspect ratio
        if w > h:
            w, h = 256, max(h*256/w, 1)
        else:
            w, h = max(w*256/h, 1), 256
        thumb_layer.scale(w, h, False)
    if thumb.get_precision() != Gimp.Precision.U8_NON_LINEAR:
        thumb.convert_precision (Gimp.Precision.U8_NON_LINEAR)
    store_layer(thumb, thumb_layer, 'Thumbnails/thumbnail.png')
    thumb.delete()

    # write stack.xml
    xml = ET.tostring(xml_image, encoding='UTF-8')
    write_file_str(orafile, 'stack.xml', xml)

    # finish up
    orafile.close()
    os.rmdir(tempdir)
    if os.path.exists(file.peek_path()):
        os.remove(file.peek_path()) # win32 needs that
    os.rename(file.peek_path() + '.tmpsave', file.peek_path())

    Gimp.progress_end()

    return Gimp.ValueArray.new_from_values([
        GObject.Value(Gimp.PDBStatusType, Gimp.PDBStatusType.SUCCESS)
    ])

def load_ora(procedure, run_mode, file, metadata, flags, config, data):
    tempdir = tempfile.mkdtemp('gimp-plugin-file-openraster')
    orafile = zipfile.ZipFile(file.peek_path())
    stack, w, h = get_image_attributes(orafile)

    Gimp.progress_init("Loading openraster image")

    img = Gimp.Image.new(w, h, Gimp.ImageBaseType.RGB)

    def get_layers(root):
        """iterates over layers and nested stacks"""
        for item in root:
            if item.tag == 'layer':
                yield item
            elif item.tag == 'stack':
                yield item
                for subitem in get_layers(item):
                    yield subitem
                yield NESTED_STACK_END

    parent_groups = []

    # Number of top level layers for tracking progress
    lay_cnt = len(stack)

    layer_no = 0
    selected_layers = []
    for item in get_layers(stack):
        prev_lay = layer_no

        if item is NESTED_STACK_END:
            parent_groups.pop()
            continue

        if item.tag == 'stack':
            name, x, y, opac, visible, layer_mode, selected, locked = get_group_layer_attributes(item)
            gimp_layer = Gimp.GroupLayer.new(img, name)

        else:
            path, name, x, y, opac, visible, layer_mode, selected, locked = get_layer_attributes(item)

            if not path.lower().endswith('.png'):
                continue
            if not name:
                # use the filename without extension as name
                n = os.path.basename(path)
                name = os.path.splitext(n)[0]

            # create temp file. Needed because gimp cannot load files from inside a zip file
            tmp = os.path.join(tempdir, 'tmp.png')
            with open(tmp, 'wb') as fid:
                try:
                    data = orafile.read(path)
                except KeyError:
                    # support for bad zip files (saved by old versions of this plugin)
                    data = orafile.read(path.encode('utf-8'))
                    print('WARNING: bad OpenRaster ZIP file. There is an utf-8 encoded filename that does not have the utf-8 flag set:', repr(path))
                fid.write(data)

            # import layer, set attributes and add to image
            pdb_proc   = Gimp.get_pdb().lookup_procedure('gimp-file-load-layer')
            pdb_config = pdb_proc.create_config()
            pdb_config.set_property('run-mode', Gimp.RunMode.NONINTERACTIVE)
            pdb_config.set_property('image', img)
            pdb_config.set_property('file', Gio.File.new_for_path(tmp))
            result = pdb_proc.run(pdb_config)
            if (result.index(0) == Gimp.PDBStatusType.SUCCESS):
                gimp_layer = result.index(1)
                os.remove(tmp)
            else:
                print("Error loading layer from openraster image.")

        gimp_layer.set_name(name)
        gimp_layer.set_mode(layer_mode)
        gimp_layer.set_offsets(x, y)  # move to correct position
        gimp_layer.set_opacity(opac * 100)  # a float between 0 and 100
        gimp_layer.set_visible(visible)
        gimp_layer.set_lock_content(locked)
        if selected:
            selected_layers.append(gimp_layer)

        img.insert_layer(gimp_layer,
                         parent_groups[-1][0] if parent_groups else None,
                         parent_groups[-1][1] if parent_groups else layer_no)
        if parent_groups:
            parent_groups[-1][1] += 1
        else:
            layer_no += 1

        if gimp_layer.is_group():
            parent_groups.append([gimp_layer, 0])

        if (layer_no > prev_lay):
            Gimp.progress_update(layer_no/lay_cnt)

    if len(selected_layers) > 0:
        img.set_selected_layers(selected_layers)

    Gimp.progress_end()

    os.rmdir(tempdir)

    return Gimp.ValueArray.new_from_values([
        GObject.Value(Gimp.PDBStatusType, Gimp.PDBStatusType.SUCCESS),
        GObject.Value(Gimp.Image, img),
    ]), flags


class FileOpenRaster (Gimp.PlugIn):
    ## GimpPlugIn virtual methods ##
    def do_set_i18n(self, procname):
        return True, 'gimp30-python', None

    def do_query_procedures(self):
        return [ 'file-openraster-load-thumb',
                 'file-openraster-load',
                 'file-openraster-export' ]

    def do_create_procedure(self, name):
        if name == 'file-openraster-export':
            procedure = Gimp.ExportProcedure.new(self, name,
                                                 Gimp.PDBProcType.PLUGIN,
                                                 False, export_ora, None)
            procedure.set_image_types("*");
            procedure.set_documentation (_('export an OpenRaster (.ora) file'),
                                         _('export an OpenRaster (.ora) file'),
                                         name)
            procedure.set_menu_label(_('OpenRaster'))
            procedure.set_extensions ("ora");
        elif name == 'file-openraster-load':
            procedure = Gimp.LoadProcedure.new (self, name,
                                                Gimp.PDBProcType.PLUGIN,
                                                load_ora, None)
            procedure.set_menu_label(_('OpenRaster'))
            procedure.set_documentation (_('load an OpenRaster (.ora) file'),
                                         _('load an OpenRaster (.ora) file'),
                                         name)
            procedure.set_mime_types ("image/openraster");
            procedure.set_extensions ("ora");
            procedure.set_thumbnail_loader ('file-openraster-load-thumb');
        else: # 'file-openraster-load-thumb'
            procedure = Gimp.ThumbnailProcedure.new (self, name,
                                                     Gimp.PDBProcType.PLUGIN,
                                                     thumbnail_ora, None)
            procedure.set_documentation (_('loads a thumbnail from an OpenRaster (.ora) file'),
                                         _('loads a thumbnail from an OpenRaster (.ora) file'),
                                         name)
        procedure.set_attribution('Jon Nordby', #author
                                  'Jon Nordby', #copyright
                                  '2009') #year
        return procedure

Gimp.main(FileOpenRaster.__gtype__, sys.argv)



#!/usr/bin/env python2

"""
module-dependencies.py -- GIMP library and core module dependency constructor
Copyright (C) 2010  Martin Nordholts <martinn@src.gnome.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.



This program uses graphviz (you need PyGraphViz) to construct a graph
with dependencies between GIMP library and core modules. Run it from
the source root. Note that you'll either need the very latest
PyGraphViz binding or use this hack in agraph.py:

--- agraph.py.orig        2010-01-04 16:07:46.000000000 +0100
+++ agraph.py        2010-01-04 16:13:54.000000000 +0100
@@ -1154,7 +1154,8 @@ class AGraph(object):
             raise IOError("".join(errors))

         if len(errors)>0:
-            raise IOError("".join(errors))
+            # Workaround exception throwing due to warning about cycles
+            pass
         return "".join(data)


"""


from __future__ import with_statement
from sets import Set
import os, re, pygraphviz



# First make a sanity check
if not os.path.exists("app") or not os.path.exists("libgimp"):
    print("Must be run in source root!")
    exit(-1);


# This file lives in libgimp and is used by many low-level
# libs. Exclude it in the calculations so the graph become nicer
ignored_interface_files = [
    "libgimp/libgimp-intl.h",
    ]

# List of library modules
libmodules = [
    "libgimp",
    "libgimpbase",
    "libgimpcolor",
    "libgimpconfig",
    "libgimpmath",
    "libgimpmodule",
    "libgimpthumb",
    "libgimpwidgets",
    ]

# List of app modules
# XXX: Maybe group some of these together to simplify graph?
appmodules = [
    "actions",
    "base",
    "composite",
    "config",
    "core",
    "dialogs",
    "display",
    "file",
    "gegl",
    "gui",
    "menus",
    "paint",
    "paint-funcs",
    "pdb",
    "plug-in",
    "tests",
    "text",
    "tools",
    "vectors",
    "widgets",
    "xcf",
    ]

# Bootstrap modules, i.e. modules we assume exist even though we don't
# have the code for them
boostrap_modules = [
    [ "GLib", ["glib.h"] ],
    [ "GTK+", ["gtk/gtk.h"] ],
    [ "GEGL", ["gegl.h"] ],
    [ "Pango", ["pango/pango.h"] ],
    [ "Cairo", ["cairo.h"] ],
    ]

##
# Function to determine if a filename is for an interface file
def is_interface_file(filename):
    return re.search("\.h$", filename)

##
# Function to determine if a filename is for an implementation file,
# i.e. a file that contains references to interface files
def is_implementation_file(filename):
    return re.search("\.c$", filename)

##
# Represents a software module. Think of it as a node in the
# dependency graph
class Module:
    def __init__(self, name, color, interface_files=[]):
        self.name = name
        self.color = color
        self.interface_files = Set(interface_files.__iter__())
        self.interface_file_dependencies = Set()
        self.dependencies = Set()

    def __repr__(self):
        return self.name

    def get_color(self):
        return self.color

    def get_interface_files(self):
        return self.interface_files

    def get_interface_file_dependencies(self):
        return self.interface_file_dependencies

    def get_dependencies(self):
        return self.dependencies

    def add_module_dependency(self, module):
        if self != module:
            self.dependencies.add(module)


# Represents a software module constructed from actual source code
class CodeModule(Module):
    def __init__(self, path, color):
        Module.__init__(self, path, color)

        all_files = os.listdir(path)

        # Collect interfaces this module provides
        for interface_file in filter(is_interface_file, all_files):
            self.interface_files.add(os.path.join(path, interface_file))

        # Collect dependencies to interfaces
        for filename in filter(is_implementation_file, all_files):
            with open(os.path.join(path, filename), 'r') as f:
                for line in f:
                    m = re.search("#include +[\"<](.*)[\">]", line)
                    if m:
                        interface_file = m.group(1)
                        # If the interface file appears to come from a core
                        # module, prepend with "app/"
                        m = re.search ("(.*)/.*", interface_file)
                        if m:
                            dirname = m.group(1)
                            if appmodules.__contains__(dirname):
                                interface_file = "app/" + interface_file

                        self.interface_file_dependencies.add(interface_file)

        for ignored_interface_file in ignored_interface_files:
            self.interface_file_dependencies.discard(ignored_interface_file)



# Initialize the modules to use for the dependency analysis
modules = Set()
for bootstrap_module in boostrap_modules:
    modules.add(Module(bootstrap_module[0], 'lightblue', bootstrap_module[1]))
for module_path in libmodules:
    modules.add(CodeModule(module_path, 'coral1'))
for module_path in appmodules:
    modules.add(CodeModule("app/" + module_path, 'lawngreen'))


# Map the interface files in the modules to the module that hosts them
interface_file_to_module = {}
for module in modules:
    for interface_file in module.get_interface_files():
        interface_file_to_module[interface_file] = module

# Figure out dependencies between modules
unknown_interface_files = Set()
for module in modules:
    interface_files = filter (is_interface_file, module.get_interface_file_dependencies())
    for interface_file in interface_files:
        if interface_file_to_module.has_key(interface_file):
            module.add_module_dependency(interface_file_to_module[interface_file])
        else:
            unknown_interface_files.add(interface_file)
if False:
    print "Unknown interface files:", unknown_interface_files

# Construct a GraphViz graph from the modules
dependency_graph = pygraphviz.AGraph(directed=True)
for module in modules:
    dependency_graph.add_node(module, fillcolor=module.get_color(), style='filled')
for module in modules:
    for depends_on in module.get_dependencies():
        dependency_graph.add_edge(module, depends_on)

# If module A depends on module B, and module B depends on module C, A
# gets C implicitly. Perform a transitive reduction on the graph to
# reflect this
dependency_graph.tred()

# Write result
if True:
    dependency_graph.draw("devel-docs/gimp-module-dependencies.svg", prog="dot")
else:
    dependency_graph.write("devel-docs/gimp-module-dependencies.dot")



#!/usr/bin/env python3
import sys
import os
import subprocess
import filecmp
import shutil
from datetime import datetime

# This is a wrapper to the tools/gimp-mkenums perl script which:
# * sets a few common values;
# * updates the ${filebase}enums.c directly in the source directory in
#   order for it to be versioned.
# * Create a no-op stamp-header file to be included by the resulting
#   enums.c. The goal is to make sure that meson will trigger a rebuild
#   of the enums.c generation before compiling, if the enums.h changed.
#   See the explanation here:
#   https://github.com/mesonbuild/meson/issues/10196#issuecomment-1080742592
#   This is also the trick used for pdbgen.

# Arguments to this script:
# The perl binary to use.
PERL = sys.argv[1]
# Root of the source directory.
top_srcdir = sys.argv[2]
# Current source folder.
srcdir = sys.argv[3]
# Current build folder.
builddir = sys.argv[4]
# Base of the generated enums.c file name.
filebase = sys.argv[5]
# Includes before #include "${filebase}enums.h"
preincludes = sys.argv[6]
# Includes after #include "${filebase}enums.h"
postincludes = sys.argv[7]
# Value for --dtail option if the default doesn't fit.
dtail = sys.argv[8] if len(sys.argv) >= 9 else None

if not dtail:
  dtail = "    { 0, NULL, NULL }\n  };\n\n  static GType type = 0;\n\n  if (G_UNLIKELY (! type))\n    {\n      type = g_@type@_register_static (\"@EnumName@\", values);\n      gimp_type_set_translation_context (type, \"@enumnick@\");\n      gimp_@type@_set_value_descriptions (type, descs);\n    }\n\n  return type;\n}\n"

args = [
  PERL, os.path.join(top_srcdir, 'tools', 'gimp-mkenums'),
  '--fhead', f'#include "stamp-{filebase}enums.h"\n' f'#include "config.h"\n' f'{preincludes}' f'#include "{filebase}enums.h"\n' f'{postincludes}',
  '--fprod', '\n/* enumerations from "@basename@" */',
  '--vhead', 'GType\n' '@enum_name@_get_type (void)\n' '{\n' '  static const G@Type@Value values[] =\n' '  {',
  '--vprod', '    { @VALUENAME@, "@VALUENAME@", "@valuenick@" },',
  '--vtail', '    { 0, NULL, NULL }\n  };\n',
  '--dhead', '  static const Gimp@Type@Desc descs[] =\n  {',
  '--dprod', '    { @VALUENAME@, @valuedesc@, @valuehelp@ },' '@if (\'@valueabbrev@\' ne \'NULL\')@\n' '    /* Translators: this is an abbreviated version of @valueudesc@.\n' '       Keep it short. */\n' '    { @VALUENAME@, @valueabbrev@, NULL },@endif@',
  '--dtail', dtail,
  os.path.join(f"{filebase}enums.h")
]
tmp_enums_c = os.path.join(builddir, f"{filebase}enums-tmp.c")
with open(tmp_enums_c, 'w') as f:
  subprocess.run(args, check=True, stdout=f, text=True, cwd=srcdir)

target_enums_c = os.path.join(srcdir, f"{filebase}enums.c")
if not filecmp.cmp(tmp_enums_c, target_enums_c, shallow=False):
  shutil.copy(tmp_enums_c, target_enums_c)
else:
  try:
    os.utime(target_enums_c, None)
  except OSError:
    pass

with open(os.path.join(builddir, f"stamp-{filebase}enums.h"), 'w') as f:
  f.write(f"/* Generated on {datetime.now().strftime('%a %b %d %H:%M:%S %Z %Y')}. */\n")



#!/usr/bin/python3
import platform
import os
import subprocess
import sys
import re
try:
  import charset_normalizer
except ImportError:
  result = subprocess.run([sys.executable, '--version', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
  output = result.stdout + result.stderr
  if re.search(r'\bMSC\b', output, re.IGNORECASE):
    subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'charset_normalizer'])
  else:
    MINGW_PACKAGE_PREFIX = "mingw-w64-clang-aarch64" if platform.machine() == "ARM64" else "mingw-w64-clang-x86_64"
    subprocess.check_call(['powershell', os.getenv("MSYS_ROOT") + '/usr/bin/pacman', '--noconfirm', '-S', '--needed', f"{MINGW_PACKAGE_PREFIX}-python-charset-normalizer"])
finally:
  from charset_normalizer import detect

langfilePath = sys.argv[1]
AppVer = sys.argv[2]

# Detect the encoding of the file
with open(langfilePath, 'rb') as file:
  raw_data = file.read()
detected_encoding = detect(raw_data)['encoding']
print(f"(INFO): temporarily patching {detected_encoding} {langfilePath} with {AppVer}")

# Read the file content with detected encoding
with open(langfilePath, 'r', encoding=detected_encoding) as file:
  lines = file.readlines()

# Patch 'SetupWindowTitle' and 'UninstallAppFullTitle'
for i, line in enumerate(lines):
  if 'SetupWindowTitle' in line:
    before = line.strip()
    after = re.sub(r'%1', f'%1 {AppVer}', before)
    lines[i] = line.replace(before, after)
  if 'UninstallAppFullTitle' in line:
    before = line.strip()
    after = re.sub(r'%1', f'%1 {AppVer}', before)
    lines[i] = line.replace(before, after)

# Write the patched content back to the file using the detected encoding
with open(langfilePath, 'w', encoding=detected_encoding) as file:
  file.writelines(lines)



#!/usr/bin/env python3

"""
performance-log-coalesce.py -- Coalesce GIMP performance log symbols, by
                               filling-in missing base addresses
Copyright (C) 2018  Ell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.


Usage: performance-log-coalesce.py < infile > outfile
"""

from xml.etree import ElementTree
import sys

empty_element = ElementTree.Element ("")

# Read performance log from STDIN
log = ElementTree.fromstring (sys.stdin.buffer.read ())

address_map = log.find ("address-map")

if address_map is not None:
    addresses = []

    # Create base addresses dictionary
    base_addresses = {}

    for address in address_map.iterfind ("address"):
        symbol = address.find ("symbol").text
        source = address.find ("source").text
        base   = address.find ("base").text

        if symbol and source and not base:
            key   = (symbol, source)
            value = int (address.get ("value"), 0)

            base_addresses[key] = min (value, base_addresses.get (key, value))

            addresses.append (address)

    # Fill-in missing base addresses
    for address in addresses:
        symbol = address.find ("symbol").text
        source = address.find ("source").text

        address.find ("base").text = hex (base_addresses[(symbol, source)])

# Write performance log to STDOUT
sys.stdout.buffer.write (ElementTree.tostring (log))



#!/usr/bin/env python3
import os
import sys
import subprocess
from pathlib import Path

GIMP_EXE = sys.argv[1]
TEST_FILE = sys.argv[2]
SRC_DIR = os.path.dirname(TEST_FILE)
SRC_DIR = Path(os.path.realpath(SRC_DIR)).resolve().as_posix()
PYGIMP_DIR = os.path.dirname(__file__)
PYGIMP_DIR = Path(os.path.realpath(PYGIMP_DIR)).resolve().as_posix()

if not os.path.isfile(TEST_FILE):
  print(f"ERROR: file '{TEST_FILE}' does not exist!")
  sys.exit(1)

with open(TEST_FILE, 'r') as f:
  first_char = f.read(1)

if first_char != '#':
  # Note: I don't actually care that it's a shebang, just that it's a comment,
  # hence a useless line, because I'm going to remove it and replace it with
  # gimp_assert() import line.
  # This will make much easier to debug tests with meaningful line numbers.
  print(f"ERROR: file '{TEST_FILE}' should start with a shebang: #!/usr/bin/env python3")
  sys.exit(1)

header = f"""import os; import sys; sys.path.insert(0, '{PYGIMP_DIR}'); from pygimp.utils import gimp_assert;
import pygimp.utils; pygimp.utils.gimp_test_filename = '{TEST_FILE}'"""

with open(TEST_FILE, 'r') as f:
  content = f.readlines()
cmd = header + ''.join(content[1:])
subprocess.run([sys.executable, GIMP_EXE, "-nis", "--batch-interpreter", "python-fu-eval", "-b", cmd, "--quit"], check=True)



def my_signal_handler (event):
  if (isinstance(event, gdb.SignalEvent)):
    gdb.write("Eeeeeeeeeeeek: in-build GIMP crashed!\n")
    gdb.execute('info threads')
    gdb.execute("thread apply all backtrace full")

gdb.events.stop.connect(my_signal_handler)
gdb.execute("run")



#!/usr/bin/env python3
import os
import sys
import subprocess
from pathlib import Path

GIMP_EXE = sys.argv[1]
TEST_FILE = sys.argv[2]
SRC_DIR = os.path.dirname(TEST_FILE)
SRC_DIR = Path(os.path.realpath(SRC_DIR)).resolve().as_posix()
TEST_NAME = sys.argv[3]

cmd = (
  f"import os; import sys; sys.path.insert(0, '{SRC_DIR}'); from pygimp.utils import gimp_c_assert; "
  f"proc = Gimp.get_pdb().lookup_procedure('{TEST_NAME}'); gimp_c_assert('{TEST_FILE}', 'Test PDB procedure does not exist: {{}}'.format('{TEST_NAME}'), proc is not None); "
  f"result = proc.run(None); "
  f"print('SUCCESS') if result.index(0) == Gimp.PDBStatusType.SUCCESS else gimp_c_assert('{TEST_FILE}', result.index(1), False);"
)
subprocess.run([sys.executable, GIMP_EXE, "-nis", "--batch-interpreter", "python-fu-eval", "-b", cmd, "--quit"], check=True)



#!/usr/bin/env python3

EPSILON=1e-6

def DBL(c):
  return c / 255.0

samples = [
  { 'str': "#000000",
    'fail': False,
    'red': 0.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 1.0 },
  { 'str': "#FFff00",
    'fail': False,
    'red': 1.0,
    'green': 1.0,
    'blue': 0.0,
    'alpha': 1.0 },
  { 'str': "#6495ed",
    'fail': False,
    'red': DBL(100),
    'green': DBL(149),
    'blue': DBL(237),
    'alpha': 1.0 },
  { 'str': "#fff",
    'fail': False,
    'red': 1.0,
    'green': 1.0,
    'blue': 1.0,
    'alpha': 1.0 },
  #{ 'str': "#64649595eded",
    #'fail': False,
    #'red': 1.0,
    #'green': 1.0,
    #'blue': 0.0,
    #'alpha': 1.0 },
  { 'str': "rgb(0,0,0)",
    'fail': False,
    'red': 0.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 1.0 },
  { 'str': "rgb(100,149,237)",
    'fail': False,
    'red': DBL(100),
    'green': DBL(149),
    'blue': DBL(237),
    'alpha': 1.0 },
  { 'str': "rgba(100%,0,100%,0.5)",
    'fail': False,
    'red': 1.0,
    'green': 0.0,
    'blue': 1.0,
    'alpha': 0.5 },
  { 'str': "rgb(100%,149,20%)",
    'fail': False,
    'red': 1.0,
    'green': DBL(149),
    'blue': 0.2,
    'alpha': 1.0 },
  { 'str': "rgb(foobar)",
    'fail': True,
    'red': 0.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 1.0 },
  { 'str': "rgb(100,149,237",
    'fail': True,
    'red': 0.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 1.0 },
  { 'str': "rED",
    'fail': False,
    'red': 1.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 1.0 },
  { 'str': "cornflowerblue",
    'fail': False,
    'red': DBL(100),
    'green': DBL(149),
    'blue': DBL(237),
    'alpha': 1.0 },
  { 'str': "    red",
    'fail': False,
    'red': 1.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 1.0 },
  { 'str': "red      ",
    'fail': False,
    'red': 1.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 1.0 },
  { 'str': "red",
    'fail': False,
    'red': 1.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 1.0 },
  { 'str': "red  blue",
    'fail': True,
    'red': 0.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 0.0 },
  { 'str': "transparent",
    'fail': False,
    'red': 0.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 0.0 },
  { 'str': "23foobar",
    'fail': True,
    'red': 0.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 0.0 },
  { 'str': "",
    'fail': True,
    'red': 0.0,
    'green': 0.0,
    'blue': 0.0,
    'alpha': 0.0 }
]

for sample in samples:
  color = Gimp.color_parse_css(sample['str'])
  print(color)
  if sample['fail']:
    gimp_assert('Parsing "{}" should fail.'.format(sample['str']), color is None)
  else:
    gimp_assert('Parsing "{}" should generate a color.'.format(sample['str']), color is not None)

    r, g, b, a = color.get_rgba_with_space(Babl.format("R'G'B'A double"))
    gimp_assert('"{}" generated ({}, {}, {}, {}) - expected ({}, {}, {}, {}).'.format(sample['str'],
                                                                                      r, g, b, a,
                                                                                      sample['red'],
                                                                                      sample['green'],
                                                                                      sample['blue'],
                                                                                      sample['alpha']),
                abs(r - sample['red']) < EPSILON and abs(g - sample['green']) < EPSILON and
                abs(b - sample['blue']) < EPSILON and abs(a - sample['alpha']) < EPSILON)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


# For a detailed explanation of the parameters of this plugin, see :
# https://gitlab.gnome.org/GNOME/gimp/-/issues/4368#note_763460

# little known, colorsys is part of Python's stdlib
from colorsys import rgb_to_yiq
from textwrap import dedent
from random import randint
import struct

import gi
gi.require_version('Babl', '0.1')
from gi.repository import Babl
gi.require_version('Gimp', '3.0')
from gi.repository import Gimp
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gio
import sys

def N_(message): return message
def _(message): return GLib.dgettext(None, message)


AVAILABLE_CHANNELS = (_("Red"), _("Green"), _("Blue"),
                      _("Luma (Y)"),
                      _("Hue"), _("Saturation"), _("Value"),
                      _("Saturation (HSL)"), _("Lightness (HSL)"),
                      _("Index"),
                      _("Random"))

channel_getters = [
    (lambda v, i: v.get_rgba()[0]),
    (lambda v, i: v.get_rgba()[1]),
    (lambda v, i: v.get_rgba()[2]),

    (lambda v, i: rgb_to_yiq(v.get_rgba()[0], v.get_rgba()[1], v.get_rgba()[1])[0]),

    # TODO: Replace with gegl_color_get_hsv () when available in Gegl
    (lambda v, i: gegl_color_bytes_convert (v, "HSV float", 'fff', 0)),
    (lambda v, i: gegl_color_bytes_convert (v, "HSV float", 'fff', 1)),
    (lambda v, i: gegl_color_bytes_convert (v, "HSV float", 'fff', 2)),

    (lambda v, i: gegl_color_bytes_convert (v, "HSL float", 'fff', 1)),
    (lambda v, i: gegl_color_bytes_convert (v, "HSL float", 'fff', 2)),

    (lambda v, i: i),
    (lambda v, i: randint(0, 0x7fffffff)),

    (lambda v, i: gegl_color_bytes_convert (v, "CIE Lab float", 'fff', 0)),
    (lambda v, i: gegl_color_bytes_convert (v, "CIE Lab float", 'fff', 1)),
    (lambda v, i: gegl_color_bytes_convert (v, "CIE Lab float", 'fff', 2)),

    (lambda v, i: gegl_color_bytes_convert (v, "CIE LCH(ab) float", 'fff', 1)),
    (lambda v, i: gegl_color_bytes_convert (v, "CIE LCH(ab) float", 'fff', 2))
]

GRAIN_SCALE = (1.0, 1.0 , 1.0,
              1.0,
              360., 100., 100.,
              100., 100.,
              16384.,
              float(0x7ffffff),
              100., 256., 256.,
              256., 360.,)

def gegl_color_bytes_convert (color, format, precision, index):
    color_bytes = color.get_bytes(Babl.format(format))
    data = color_bytes.get_data()
    result = struct.unpack (precision, data)

    return result[index]

slice_expr_doc = N_("""
    Format is 'start:nrows,length' . All items are optional.

    The empty string selects all items, as does ':'
    ':4,' makes a 4-row selection out of all colors (length auto-determined)
    ':4' also.
    ':1,4' selects the first 4 colors
    ':,4' selects rows of 4 colors (nrows auto-determined)
    ':3,4' selects 3 rows of 4 colors
    '4:' selects a single row of all colors after 4, inclusive.
    '3:,4' selects rows of 4 colors, starting at 3 (nrows auto-determined)
    '2:3,4' selects 3 rows of 4 colors (12 colors total), beginning at index 2.
    '4' is illegal (ambiguous)
""")


def parse_slice(s, numcolors):
    """Parse a slice spec and return (start, nrows, length)
    All items are optional. Omitting them makes the largest possible selection that
    exactly fits the other items.

    In general, slices are comparable to a numpy sub-array.
    'start at element START, with shape (NROWS, LENGTH)'

    """
    s = s.strip()

    def notunderstood():
        raise ValueError('Slice %r not understood. Should be in format'
                         ' START?:NROWS?,ROWLENGTH? eg. "0:4,16".' % s)
    def _int(v):
        try:
            return int(v)
        except ValueError:
            notunderstood()
    if s in ('', ':', ':,'):
        return 0, 1, numcolors  # entire palette, one row
    if s.count(':') != 1:
        notunderstood()
    rowpos = s.find(':')
    start = 0
    if rowpos > 0:
        start = _int(s[:rowpos])
    numcolors -= start
    nrows = 1
    if ',' in s:
        commapos = s.find(',')
        nrows = s[rowpos+1:commapos]
        length = s[commapos+1:]
        if not nrows:
            if not length:
                notunderstood()
            else:
                length = _int(length)
                if length == 0:
                    notunderstood()
                nrows = numcolors // length
                if numcolors % length:
                    nrows = -nrows
        elif not length:
            nrows = _int(nrows)
            if nrows == 0:
                notunderstood()
            length = numcolors // nrows
            if numcolors % nrows:
                length = -length
        else:
            nrows = _int(nrows)
            if nrows == 0:
                notunderstood()
            length = _int(length)
            if length == 0:
                notunderstood()
    else:
        nrows = _int(s[rowpos+1:])
        if nrows == 0:
            notunderstood()
        length = numcolors // nrows
        if numcolors % nrows:
            length = -length
    return start, nrows, length


def quantization_grain(channel, g):
    "Given a channel and a quantization, return the size of a quantization grain"
    g = max(1.0, g)
    if g <= 1.0:
        g = 0.00001
    else:
        g = max(0.00001, GRAIN_SCALE[channel] / g)
    return g


def palette_sort(palette, selection, slice_expr, channel1, ascending1,
                 channel2, ascending2, quantize, pchannel, pquantize):

    grain1 = quantization_grain(channel1, quantize)
    grain2 = quantization_grain(channel2, quantize)
    pgrain = quantization_grain(pchannel, pquantize)

    # If palette is read only, work on a copy:
    editable = palette.is_editable()
    if not editable:
        palette = palette.duplicate()

    num_colors = palette.get_color_count()

    start, nrows, length = None, None, None
    if selection == "auto-slice":
        def find_index(color, startindex=0):
            for i in range(startindex, num_colors):
                rgba = palette.get_entry_color(i)

                if hexcolor(rgba) == hexcolor(color):
                    return i
            return None
        def hexcolor(c):
            rgba = c.get_rgba()
            return "#%02x%02x%02x" % (int(255 * rgba[0]), int(255 * rgba[1]), int(255 * rgba[2]))
        fg = Gimp.context_get_foreground()
        bg = Gimp.context_get_background()
        start = find_index(fg)
        end = find_index(bg)

        if start is None:
            raise ValueError("Couldn't find foreground color %s in palette" % hexcolor(fg))
        if end is None:
            raise ValueError("Couldn't find background color %s in palette" % hexcolor(bg))
        if find_index(fg, start + 1):
            raise ValueError('Autoslice cannot be used when more than one'
                             ' instance of an endpoint'
                             ' (%s) is present' % hexcolor(fg))
        if find_index(bg, end + 1):
            raise ValueError('Autoslice cannot be used when more than one'
                             ' instance of an endpoint'
                             ' (%s) is present' % hexcolor(bg))
        if start > end:
            end, start = start, end
        length = (end - start) + 1
        try:
            _, nrows, _ = parse_slice(slice_expr, length)
            nrows = abs(nrows)
            if length % nrows:
                raise ValueError('Total length %d not evenly divisible'
                                 ' by number of rows %d' % (length, nrows))
            length //= nrows
        except ValueError:
            # bad expression is okay here, just assume one row
            nrows = 1
        # remaining behavior is implemented by "slice-array" 'inheritance'.
        selection = "slice-array"
    elif selection in ("slice-array", "partitioned"):
        start, nrows, length = parse_slice(slice_expr, num_colors)

    channels_getter_1 = channel_getters[channel1]
    channels_getter_2 = channel_getters[channel2]

    def get_colors(start, end):
        result = []
        for i in range(start, end):
            entry =  (palette.get_entry_name(i)[1],
                      palette.get_entry_color(i))
            index1 = channels_getter_1(entry[1], i)
            index2 = channels_getter_2(entry[1], i)
            index = ((index1 - (index1 % grain1)) * (1 if ascending1 else -1),
                     (index2 - (index2 % grain2)) * (1 if ascending2 else -1)
                    )
            result.append((index, entry))
        return result

    if selection == "all":
        entry_list = get_colors(0, num_colors)
        entry_list.sort(key=lambda v: v[0])
        for i in range(num_colors):
            palette.set_entry_name(i, entry_list[i][1][0])
            palette.set_entry_color(i, entry_list[i][1][1])

    elif selection == "partitioned":
        if num_colors < (start + length * nrows) - 1:
            raise ValueError('Not enough entries in palette to '
                             'sort complete rows! Got %d, expected >=%d' %
                             (num_colors, start + length * nrows))
        pchannels_getter = channel_getters[pchannel]
        for row in range(nrows):
            partition_spans = [1]
            rowstart = start + (row * length)
            old_color = palette.get_entry_color(rowstart)
            old_partition = pchannels_getter(old_color, rowstart)
            old_partition = old_partition - (old_partition % pgrain)
            for i in range(rowstart + 1, rowstart + length):
                this_color = palette.get_entry_color(i)
                this_partition = pchannels_getter(this_color, i)
                this_partition = this_partition - (this_partition % pgrain)
                if this_partition == old_partition:
                    partition_spans[-1] += 1
                else:
                    partition_spans.append(1)
                    old_partition = this_partition
            base = rowstart
            for size in partition_spans:
                palette_sort(palette, "slice-array", '%d:1,%d' % (base, size),
                             channel1, ascending1,
                             channel2, ascending2,
                             quantize, 0, 1.0)
                base += size
    else:
        # "slice-array" and "auto-slice"
        stride = length
        if num_colors < (start + stride * nrows) - 1:
            raise ValueError('Not enough entries in palette to sort '
                             'complete rows! Got %d, expected >=%d' %
                             (num_colors, start + stride * nrows))

        for row_start in range(start, start + stride * nrows, stride):
            sublist = get_colors(row_start, row_start + stride)
            sublist.sort(key=lambda v: v[0])
            for i, entry in zip(range(row_start, row_start + stride), sublist):
                palette.set_entry_name(i, entry[1][0])
                palette.set_entry_color(i, entry[1][1])

    return palette

# FIXME: Write humanly readable help -
# See for reference: https://gitlab.gnome.org/GNOME/gimp/-/issues/4368#note_763460

# Important to describe the general effect on palettes rather than details of the sort.
help_doc = r"""
Sorts a palette, or part of a palette.
Sorts the given palette when it is editable, otherwise creates a new sorted palette.
The default is a 1D sort, but you can also sort over two color channels
or create a 2D sorted palette with sorted rows.
You can optionally install colormath (https://pypi.python.org/pypi/colormath/1.0.8)
to GIMP's Python to get even more channels to choose from.
"""

class PaletteSort (Gimp.PlugIn):
    ## GimpPlugIn virtual methods ##
    def do_set_i18n(self, procname):
        return True, 'gimp30-python', None

    def do_query_procedures(self):
        return ["python-fu-palette-sort"]

    def do_create_procedure(self, name):
        procedure = None
        if name == "python-fu-palette-sort":
            procedure = Gimp.Procedure.new(self, name,
                                           Gimp.PDBProcType.PLUGIN,
                                           self.run, None)
            procedure.set_menu_label(_("_Sort Palette..."))
            procedure.set_documentation(
                _("Sort the colors in a palette"),
                help_doc,
                ""
            )
            procedure.set_attribution("João S. O. Bueno, Carol Spears, David Gowers",
                                      "João S. O. Bueno, Carol Spears, David Gowers",
                                      "2006-2014")
            procedure.add_menu_path ('<Palettes>/Palettes Menu')

            procedure.add_enum_argument ("run-mode", _("Run mode"),
                                         _("The run mode"), Gimp.RunMode,
                                         Gimp.RunMode.INTERACTIVE,
                                         GObject.ParamFlags.READWRITE)
            procedure.add_palette_argument ("palette", _("_Palette"),
                                            _("Palette"), True,
                                            None, True, # Default to context.
                                            GObject.ParamFlags.READWRITE)

            selection_choice = Gimp.Choice.new()
            selection_choice.add("all", 0, _("All"), "")
            selection_choice.add("slice-array", 1, _("Slice / Array"), "")
            selection_choice.add("auto-slice", 2, _("Autoslice (fg->bg)"), "")
            selection_choice.add("partitioned", 3, _("Partitioned"), "")
            procedure.add_choice_argument ("selections", _("Select_ions"), _("Selections"),
                                           selection_choice, "all", GObject.ParamFlags.READWRITE)
            # TODO: It would be much simpler to replace the slice expression with three
            # separate parameters: start-index, number-of-rows, row_length
            procedure.add_string_argument ("slice-expr", _("Slice _expression"),
                                           _(slice_expr_doc), "",
                                           GObject.ParamFlags.READWRITE)
            channel_choice = Gimp.Choice.new()
            self.add_choices (channel_choice)
            procedure.add_choice_argument ("channel1", _("Channel _to sort"), _("Channel to sort"),
                                           channel_choice, "luma", GObject.ParamFlags.READWRITE)
            procedure.add_boolean_argument ("ascending1", _("_Ascending"), _("Ascending"),
                                            True, GObject.ParamFlags.READWRITE)
            channel_choice2 = Gimp.Choice.new()
            self.add_choices (channel_choice2)
            procedure.add_choice_argument ("channel2", _("Secondary C_hannel to sort"),
                                           _("Secondary Channel to sort"), channel_choice2,
                                           "saturation", GObject.ParamFlags.READWRITE)
            procedure.add_boolean_argument ("ascending2", _("Ascen_ding"), _("Ascending"),
                                            True, GObject.ParamFlags.READWRITE)
            procedure.add_double_argument ("quantize", _("_Quantization"),
                                           _("Quantization"), 0.0, 1.0, 0.0,
                                           GObject.ParamFlags.READWRITE)
            pchannel_choice = Gimp.Choice.new()
            self.add_choices (pchannel_choice)
            procedure.add_choice_argument ("pchannel", _("Partitionin_g channel"),
                                           _("Partitioning channel"), pchannel_choice,
                                           "luma", GObject.ParamFlags.READWRITE)
            procedure.add_double_argument ("pquantize", _("Partition q_uantization"),
                                           _("Partition quantization"), 0.0, 1.0, 0.0,
                                           GObject.ParamFlags.READWRITE)

            procedure.add_palette_return_value ("new-palette", _("Palette"),
                                                _("Palette"), GObject.ParamFlags.READWRITE)

        return procedure

    def run(self, procedure, config, data):
        run_mode = config.get_property("run-mode")
        palette  = config.get_property("palette")

        if palette is None or not palette.is_valid():
            if palette is not None:
              sys.stderr.write(f'Invalid palette id: {palette.get_id()}\n')
              sys.stderr.write('This should not happen. Please report to GIMP project.\n')
              sys.stderr.write('Falling back to context palette instead.\n')
            palette = Gimp.context_get_palette()
            config.set_property("palette", palette)

        if not palette.is_valid():
            palette_name = palette.get_id()
            error = f'Invalid palette id: {palette_name}'
            return procedure.new_return_values(Gimp.PDBStatusType.CALLING_ERROR,
                                               GLib.Error(error))

        if run_mode == Gimp.RunMode.INTERACTIVE:
            gi.require_version('GimpUi', '3.0')
            from gi.repository import GimpUi
            gi.require_version('Gtk', '3.0')
            from gi.repository import Gtk

            GimpUi.init('python-fu-palette-sort')
            dialog = GimpUi.ProcedureDialog(procedure=procedure, config=config)

            config.set_property("palette", None)
            dialog.fill (["palette"])
            dialog.fill (["selections","slice-expr"])
            dialog.fill (["channel1", "ascending1"])
            dialog.fill (["channel2","ascending2", "quantize"])
            dialog.fill (["pchannel","pquantize"])

            if not dialog.run():
                dialog.destroy()
                return procedure.new_return_values(Gimp.PDBStatusType.CANCEL, GLib.Error())

            dialog.destroy()

        palette    = config.get_property("palette")
        selection  = config.get_property("selections")
        slice_expr = config.get_property("slice-expr")
        channel1   = config.get_choice_id("channel1")
        ascending1 = config.get_property ("ascending1")
        channel2   = config.get_choice_id("channel2")
        ascending2 = config.get_property ("ascending2")
        quantize   = config.get_property ("quantize")
        pchannel   = config.get_choice_id("pchannel")
        pquantize  = config.get_property ("pquantize")
        try:
            new_palette = palette_sort(palette, selection, slice_expr, channel1, ascending1,
                                       channel2, ascending2, quantize, pchannel, pquantize)
        except ValueError as err:
            return procedure.new_return_values(Gimp.PDBStatusType.EXECUTION_ERROR,
                                               GLib.Error(str(err)))

        return_val = procedure.new_return_values(Gimp.PDBStatusType.SUCCESS, GLib.Error())
        value = GObject.Value(Gimp.Palette, new_palette)
        return_val.remove(1)
        return_val.insert(1, value)
        return return_val

    def add_choices (self, choice):
       #TODO: Re-incorporate LAB and LCHab options with GeglColor
       choice.add("red", 0, _("Red"), "")
       choice.add("green", 1, _("Green"), "")
       choice.add("blue", 2, _("Blue"), "")
       choice.add("luma", 3, _("Luma (Y)"), "")
       choice.add("hue", 4, _("Hue"), "")
       choice.add("saturation", 5, _("Saturation"), "")
       choice.add("value", 6, _("Value"), "")
       choice.add("saturation-hsl", 7, _("Saturation (HSL)"), "")
       choice.add("lightness-hsl", 8, _("Lightness (HSL)"), "")
       choice.add("index", 9, _("Index"), "")
       choice.add("random", 10, _("Random"), "")
       choice.add("lightness-lab", 11, _("Lightness (LAB)"), "")
       choice.add("a-color", 12, _("A-color"), "")
       choice.add("b-color", 13, _("B-color"), "")
       choice.add("chroma-lchab", 14, _("Chroma (LCHab)"), "")
       choice.add("hue-lchab", 15, _("Hue (LCHab)"), "")

Gimp.main(PaletteSort.__gtype__, sys.argv)



#!/usr/bin/env python3

try:
  # fcntl only exists on UNIX-like platforms. On particular, it won't
  # exist on Windows. But it's fine, because we don't need to
  # exclusively lock this script on Windows.
  import fcntl
except:
  pass
import os
import random
import re
import shutil
import string
import subprocess
import sys
import tempfile
from pathlib import Path

# In some case, this script may not be run concurrently, in particular
# on macOS where we need to set rpath and install_name_tool doesn't like
# when a RPATH already exists.
lock = None

def cleanup(lock):
  if lock is not None:
    fcntl.flock(lock, fcntl.LOCK_UN)
    os.close(lock)

try:
  GIMP_GLOBAL_BUILD_ROOT = os.environ.get("GIMP_GLOBAL_BUILD_ROOT", ".")

  suffix = ''.join(random.choices(string.ascii_lowercase + string.digits, k=6))
  GIMP3_DIRECTORY = os.path.join(GIMP_GLOBAL_BUILD_ROOT, f".GIMP3-build-config-{suffix}")
  os.makedirs(GIMP3_DIRECTORY, mode=0o700, exist_ok=False)
  os.environ["GIMP3_DIRECTORY"] = GIMP3_DIRECTORY
  sys.stderr.write(f"INFO: temporary GIMP configuration directory: {GIMP3_DIRECTORY}")

  # Earlier code used to set DYLD_LIBRARY_PATH environment variable instead, but
  # it didn't work on contributor's builds because of System Integrity
  # Protection (SIP), though it did work in the CI which had older macOS.
  # So, we just set LC_RPATH on binaries, but this restrict us to only one
  # target at a time. See: #14236 and gimp-data/images/logo/meson.build
  rpath_array = [f"{GIMP_GLOBAL_BUILD_ROOT}/libgimp",
                 f"{GIMP_GLOBAL_BUILD_ROOT}/libgimpbase",
                 f"{GIMP_GLOBAL_BUILD_ROOT}/libgimpcolor",
                 f"{GIMP_GLOBAL_BUILD_ROOT}/libgimpconfig",
                 f"{GIMP_GLOBAL_BUILD_ROOT}/libgimpmath",
                 f"{GIMP_GLOBAL_BUILD_ROOT}/libgimpmodule",
                 f"{GIMP_GLOBAL_BUILD_ROOT}/libgimpthumb",
                 f"{GIMP_GLOBAL_BUILD_ROOT}/libgimpwidgets"]

  if "GIMP_TEMP_UPDATE_RPATH" in os.environ:
    lock = os.open(__file__, os.O_RDONLY)
    fcntl.flock(lock, fcntl.LOCK_EX)

    for binary in os.environ["GIMP_TEMP_UPDATE_RPATH"].split(":"):
      result = subprocess.run(['otool', '-l', binary], stdout=subprocess.PIPE)
      out = result.stdout.decode('utf-8', errors='replace')
      regex = re.findall(r'path (.+?) \(offset', out)
      for new_rpath in rpath_array:
        if not new_rpath in regex:
          subprocess.run(["install_name_tool", "-add_rpath", new_rpath, binary], check=True)

  #Ensure the same python from meson.build (GIMP_PYTHON_WITH_GI) is used by plugins
  #This is needed because GIMP_PYTHON_WITH_GI can not coincide with python3 from shebang
  #(on MacPorts, there is no python3 symlink, so we would misuse Xcode python3 without GI)
  different_python=False
  if sys.platform not in ['win32', 'cygwin'] and os.environ.get("GIMP_PYTHON_WITH_GI"):
    python_symlink = shutil.which("python3")
    pygobject_found=False
    if python_symlink and not os.path.samefile(python_symlink, os.environ.get("GIMP_PYTHON_WITH_GI")):
      result = subprocess.run([python_symlink,"-c","import sys, gi; version='3.0'; sys.exit(gi.check_version(version))"], check=False)
      pygobject_found = (result.returncode == 0)
    if not python_symlink or (python_symlink and not pygobject_found):
      different_python=True
      tmp_path = os.path.join(GIMP3_DIRECTORY, "tmp_python")
      os.makedirs(tmp_path, exist_ok=True)
      tmp_symlink = os.path.join(tmp_path, "python3")
      if not os.path.exists(tmp_symlink):
        os.symlink(os.environ.get("GIMP_PYTHON_WITH_GI"), tmp_symlink)
      os.environ["PATH"] = tmp_path + os.pathsep + os.environ.get("PATH", "")

  if "GIMP_DEBUG_SELF_WRAPPER" in os.environ and shutil.which("gdb"):
    sys.stderr.write(f"RUNNING: {os.environ['GIMP_DEBUG_SELF_WRAPPER']} --batch -x {os.environ['GIMP_GLOBAL_SOURCE_ROOT']}/tools/debug-in-build-gimp.py --args {os.environ['GIMP_SELF_IN_BUILD']} {' '.join(sys.argv[1:])}")
    subprocess.run([os.environ["GIMP_DEBUG_SELF_WRAPPER"],"--return-child-result","--batch","-x",f"{os.environ['GIMP_GLOBAL_SOURCE_ROOT']}/tools/debug-in-build-gimp.py","--args", os.environ["GIMP_SELF_IN_BUILD"]] + sys.argv[1:], stdin=sys.stdin, check=True)
  else:
    sys.stderr.write(f"RUNNING: {os.environ['GIMP_SELF_IN_BUILD']} {' '.join(sys.argv[1:])}")
    subprocess.run([os.environ["GIMP_SELF_IN_BUILD"]] + sys.argv[1:],stdin=sys.stdin, check=True)

  if sys.platform not in ['win32', 'cygwin'] and different_python:
    os.environ["PATH"] = os.pathsep.join([p for p in os.environ["PATH"].split(os.pathsep) if p != tmp_path])
    shutil.rmtree(tmp_path, ignore_errors=True)

  if "GIMP_TEMP_UPDATE_RPATH" in os.environ:
    for binary in os.environ["GIMP_TEMP_UPDATE_RPATH"].split(":"):
      result = subprocess.run(['otool', '-l', binary], stdout=subprocess.PIPE)
      out = result.stdout.decode('utf-8', errors='replace')
      regex = re.findall(r'path (.+?) \(offset', out)
      for new_rpath in rpath_array:
        if new_rpath in regex:
          subprocess.run(["install_name_tool", "-delete_rpath", new_rpath, binary], check=True)
    cleanup(lock)

  # Clean-up the temporary config directory after each usage, yet making sure we
  # don't get tricked by weird redirections or anything of the sort. In particular
  # we check that this is a directory with user permission, not a symlink, and
  # that it's inside inside the project build's root.
  if "GIMP3_DIRECTORY" in os.environ and os.path.isdir(GIMP3_DIRECTORY):
    if os.path.islink(GIMP3_DIRECTORY):
      sys.stderr.write(f"ERROR: $GIMP3_DIRECTORY ({GIMP3_DIRECTORY}) should not be a symlink.")
      sys.exit(1)
    used_dir_prefix = str(Path(GIMP3_DIRECTORY).resolve().as_posix())[:-6]
    tmpl_dir_prefix = f"{Path(os.environ['GIMP_GLOBAL_BUILD_ROOT']).resolve().as_posix()}/.GIMP3-build-config-"
    if used_dir_prefix != tmpl_dir_prefix:
      sys.stderr.write(f"ERROR: $GIMP3_DIRECTORY ({GIMP3_DIRECTORY}) should be under the build directory with a specific prefix.")
      sys.stderr.write(f'       "{used_dir_prefix}" != "{tmpl_dir_prefix}"')
      sys.exit(1)
    sys.stderr.write(f"INFO: Running: shutil.rmtree({GIMP3_DIRECTORY})")
    shutil.rmtree(GIMP3_DIRECTORY)
  elif not os.access(GIMP3_DIRECTORY, os.W_OK):
    sys.stderr.write(f"ERROR: $GIMP3_DIRECTORY ({GIMP3_DIRECTORY}) does not belong to the user")
    sys.exit(1)
  else:
    sys.stderr.write(f"ERROR: $GIMP3_DIRECTORY ({GIMP3_DIRECTORY}) is not a directory")
    sys.exit(1)

except subprocess.CalledProcessError as e:
  cleanup(lock)
  sys.stderr.write(f"Command failed with exit code {e.returncode}: {e.cmd}")
  sys.exit(e.returncode)
except Exception as e:
  cleanup(lock)
  sys.stderr.write(f"Error: {str(e)}")
  sys.exit(1)



#!/usr/bin/env python3

#   Gimp-Python - allows the writing of Gimp plugins in Python.
#   Copyright (C) 2003, 2005  Manish Singh <yosh@gimp.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import string
import struct
import os.path
import sys

import gi
gi.require_version('Gimp', '3.0')
from gi.repository import Gimp
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gio

def N_(message): return message
def _(message): return GLib.dgettext(None, message)


escape_table = {
    '&': '&amp;',
    '<': '&lt;',
    '>': '&gt;',
    '"': '&quot;'
}

style_def = """body {
   width: 100%%;
   font-size: %dpx;
   background-color: #000000;
   color: #ffffff;
}
"""

preamble = """<!DOCTYPE html>
<html>
<head>
<title>CSS Color XHTML written by GIMP</title>
%s
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
</head>
<body>
<pre>
"""

postamble = """\n</pre>\n</body>\n</html>\n"""

def export_colorxhtml(procedure, run_mode, image, file, options, metadata, config, data):
    if file is None:
        error = 'No file given'
        return procedure.new_return_values(Gimp.PDBStatusType.CALLING_ERROR,
                                           GLib.Error(error))

    if run_mode == Gimp.RunMode.INTERACTIVE:
        gi.require_version('GimpUi', '3.0')
        from gi.repository import GimpUi
        gi.require_version('Gtk', '3.0')
        from gi.repository import Gtk

        GimpUi.init ("file-colorxhtml-export")

        dialog = GimpUi.ProcedureDialog.new(procedure, config, _("Save as colored HTML text..."))

        dialog.fill_frame("file-frame", "source-file", False, "aux-file")
        # Set the characters text field as not enabled when the user chooses a file
        dialog.set_sensitive("characters", True, config, "source-file", True)

        dialog.fill(["characters", "file-frame", "font-size", "separate"])

        if not dialog.run():
            return procedure.new_return_values(Gimp.PDBStatusType.CANCEL,
                                               GLib.Error())

    source_file = config.get_property("source-file")
    characters  = config.get_property("characters")
    size        = config.get_property("font-size");
    separate    = config.get_property("separate")
    aux_file    = config.get_property("aux-file")

    if source_file and aux_file:
        characters = aux_file.get_path()

    #For now, work with a single layer
    layer = image.get_layers()[0]

    width = layer.get_width()
    height = layer.get_height()
    bpp = layer.get_bpp()

    html = open(file.peek_path(), 'w')

    if separate:
        dirname, cssfile = os.path.split(file.peek_path())
        cssfile = os.path.splitext(cssfile)[0] + '.css'
        cssname = os.path.join(dirname, cssfile)

        css = open(cssname, 'w')

    if source_file:
        characters_file = open(characters, 'r')
        chars = characters_file.read()
        characters_file.close()
    else:
        chars = characters

    # Remove unprintable characters from "chars".
    # TODO: This only handles ascii files.  It would be nice to handle unicode
    # files, so this could work for any language.
    goodchars = string.digits + string.ascii_letters + string.punctuation
    badchars = ''.join(chr(i) for i in range(256) if chr(i) not in goodchars)
    allchars = str.maketrans('', '', badchars)
    chars = chars.translate(allchars)

    data = [escape_table.get(c, c) for c in chars]

    if data:
        data.reverse()
    else:
        data = list('X' * 80)

    Gimp.progress_init(_("Saving as colored XHTML"))

    style = style_def % size

    if separate:
        ss = '<link rel="stylesheet" type="text/css" href="%s" />' % cssfile
        css.write(style)
    else:
        ss = '<style type="text/css">\n%s</style>' % style

    html.write(preamble % ss)

    colors = {}
    chars = []

    for y in range(0, height):

        # The characters in "chars" will be used to draw the next row.
        # Lets fill "chars" with data so it has at least enough characters.
        while len(chars) < width:
            chars[0:0] = data

        for x in range(0, width):
            pixel_color = layer.get_pixel(x, y)
            pixel_rgb = pixel_color.get_rgba()
            pixel_tuple = (int(pixel_rgb[0] * 255),
                           int(pixel_rgb[1] * 255),
                           int(pixel_rgb[2] * 255))

            color = '%02x%02x%02x' % pixel_tuple
            style = 'background-color:black; color:#%s;' % color
            char = chars.pop()

            if separate:
                if color not in colors:
                    css.write('span.N%s { %s }\n' % (color, style))
                    colors[color] = 1

                html.write('<span class="N%s">%s</span>' % (color, char))

            else:
                html.write('<span style="%s">%s</span>' % (style, char))

        html.write('\n')

        Gimp.progress_update(y / float(height))

    html.write(postamble)

    html.close()
    if separate:
        css.close()

    return Gimp.ValueArray.new_from_values([
        GObject.Value(Gimp.PDBStatusType, Gimp.PDBStatusType.SUCCESS)
    ])


class ColorXhtml(Gimp.PlugIn):
    ## GimpPlugIn virtual methods ##
    def do_set_i18n(self, procname):
        return True, 'gimp30-python', None

    def do_query_procedures(self):
        return [ 'file-colorxhtml-export' ]

    def do_create_procedure(self, name):
        procedure = None
        if name == 'file-colorxhtml-export':
            procedure = Gimp.ExportProcedure.new(self, name,
                                                 Gimp.PDBProcType.PLUGIN,
                                                 False, export_colorxhtml, None)
            procedure.set_image_types("RGB")
            procedure.set_documentation (
                _("Save as colored HTML text"),
                "Saves the image as colored XHTML text (based on Perl version by Marc Lehmann)",
                name)
            procedure.set_menu_label(_("Colored HTML text"))
            procedure.set_attribution("Manish Singh and Carol Spears",
                                      "(c) GPL V3.0 or later",
                                      "2003")

            procedure.set_extensions ("html,xhtml");

            procedure.add_boolean_argument ("source-file",
                                            _("Rea_d characters from file"),
                                            _("Read characters from file, if true, or use text entry"),
                                            False, GObject.ParamFlags.READWRITE)
            procedure.add_string_argument ("characters", _("Charac_ters"),
                                           _("Characters that will be used as colored pixels."),
                                           "foo", GObject.ParamFlags.READWRITE)
            procedure.add_int_argument ("font-size", _("Fo_nt size in pixels"),
                                        _("Font size in pixels"), 5, 100, 10,
                                        GObject.ParamFlags.READWRITE)
            procedure.add_boolean_argument ("separate", _("_Write a separate CSS file"),
                                            _("Write a separate CSS file"),
                                            False, GObject.ParamFlags.READWRITE)
            #GUI only, used to create a widget to open a file if source-file is enabled
            procedure.add_file_aux_argument ("aux-file", _("Choose File"),
                                             "", Gimp.FileChooserAction.SAVE,
                                             False, None, GObject.ParamFlags.READWRITE)

        return procedure

Gimp.main(ColorXhtml.__gtype__, sys.argv)




"""
Python scripts to test GimpFont class of libgimp API.

See comments about usage and setup in test-resource-class.py

Setup: "Sans-serif" font is selected.

There are few methods on font.
Font has no other getter/setters and is not editable.
Tests of context_set_font and a few other class methods
are in test_resource_class.py
"""

"""
Test a known, stock font "Sans-serif"
"""
font_sans_serif = Gimp.context_get_font()
assert font_sans_serif.get_id() == "Sans-serif"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   GIMP - The GNU Image Manipulation Program
#   Copyright (C) 1995 Spencer Kimball and Peter Mattis
#
#   gimpconfig.py
#   Copyright (C) 2022-2024 Jacob Boerema
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Configuration handling for use with our testing framework."""


import os

class GimpConfig(object):
    DEFAULT_CONFIG        = "config.ini"
    DEFAULT_CONFIG_FOLDER = "/tests/"
    DEFAULT_LOGFILE       = "/tests/gimp-tests.log"
    DEFAULT_DATA_FOLDER   = "/tests/"

    def __init__(self, config_path=None, log_path=None, data_path=None):
        base_path = os.path.dirname(os.path.realpath(__file__))
        if config_path is None:
            config_path = os.getenv("GIMP_TESTS_CONFIG_FILE")
        if log_path is None:
            log_path = os.getenv("GIMP_TESTS_LOG_FILE")
        if data_path is None:
            data_path = os.getenv("GIMP_TESTS_DATA_FOLDER")

        if config_path is None:
            config_path = base_path + self.DEFAULT_CONFIG_FOLDER + self.DEFAULT_CONFIG
            #print(f"Config path for file plug-in tests: {config_path}")

        self.config_folder = os.path.dirname(config_path)
        if os.path.exists(self.config_folder):
            self.config_folder += "/"
        else:
            if self.config_folder is not None:
                print (f"Config folder '{self.config_folder}' does not exist!")
            self.config_folder = self.DEFAULT_CONFIG_FOLDER

        self.config_file = os.path.basename(config_path)
        if self.config_file is None:
            self.config_file = self.DEFAULT_CONFIG

        filepath = self.config_folder + self.config_file
        if not os.path.exists(filepath):
            raise Exception(f"Config file '{filepath}' does not exist!")

        if log_path is None:
            log_path = base_path + self.DEFAULT_LOGFILE

        filepath = os.path.dirname(log_path)
        if not os.path.exists(filepath):
            if filepath is not None:
                print (f"Log file path '{filepath}' does not exist!")
            self.log_file = self.DEFAULT_LOGFILE
        else:
            self.log_file = log_path

        # xml Junit-like results use name and path of log file, but with
        # an xml extension
        self.xml_results_file = os.path.splitext(self.log_file)[0] + '.xml'

        if data_path is None:
            data_path = base_path + self.DEFAULT_DATA_FOLDER

        self.data_folder = data_path
        if not os.path.exists(self.data_folder):
            if self.data_folder is not None:
                print (f"Data path for test images '{self.data_folder}' does not exist!")
            self.data_folder = self.DEFAULT_DATA_FOLDER



#!/usr/bin/env python3

"""
performance-log-deduce.py -- Deduce GIMP performance log thread state
Copyright (C) 2018  Ell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.


Usage: performance-log-deduce.py < infile > outfile
"""

DEDUCE_MIN_N_OCCURRENCES = 10
DEDUCE_MIN_PERCENTAGE    = 0.75

from xml.etree import ElementTree
import sys

empty_element = ElementTree.Element ("")

# Read performance log from STDIN
log = ElementTree.fromstring (sys.stdin.buffer.read ())

# Construct state histogram
address_states = {}

for sample in (log.find ("samples") if log.find ("samples") is not None else empty_element).iterfind ("sample"):
    threads = (sample.find ("backtrace") if sample.find ("backtrace") is not None else empty_element).iterfind ("thread")

    for thread in threads:
        running = int (thread.get ("running"))

        frame = thread.find ("frame")

        if frame is not None:
            address = frame.get ("address")

            states = address_states.setdefault (address, [0, 0])

            states[running] += 1

# Find maximal states
for address, states in list (address_states.items ()):
    n = sum (states)

    if n >= DEDUCE_MIN_N_OCCURRENCES:
        state = 0
        m     = states[0]

        for i in range (1, len (states)):
            if states[i] > m:
                state = i
                m     = states[i]

        percentage = m / n

        if percentage >= DEDUCE_MIN_PERCENTAGE:
            address_states[address] = state
        else:
            del address_states[address]
    else:
        del address_states[address]

# Replace thread states
for sample in (log.find ("samples") if log.find ("samples") is not None else empty_element).iterfind ("sample"):
    threads = (sample.find ("backtrace") if sample.find ("backtrace") is not None else empty_element).iterfind ("thread")

    for thread in threads:
        frame = thread.find ("frame")

        if frame is not None:
            address = frame.get ("address")

            running = address_states.get (address, None)

            if running is not None:
                thread.set ("running", str (running))

# Write performance log to STDOUT
sys.stdout.buffer.write (ElementTree.tostring (log))



#!/usr/bin/env python3
import os
import re
import shutil
import subprocess
import sys
from pathlib import Path
from glob import glob

# This script is used to create a GIMP bundle dir on Windows. A bundle
# is used as source of files for making both .exe installer and .msix package
if not os.getenv("MESON_BUILD_ROOT"):
  # Let's prevent contributors from creating broken bundles
  print("\033[31m(ERROR)\033[0m: Script called standalone. Please build GIMP targeting installer or msix creation.")
  sys.exit(1)
# Get variables from MESON_BUILD_ROOT/config.h that can be used on this script
with open("config.h") as file:
  for line in file:
    match = re.match(r'^#\s*define\s+(\S+)(?:\s+(.*))?$', line)
    if match:
      key, value = match.groups()
      if value is None or not value.strip():
        value = "1" #needed when there is no explicit value
      else:
        value = value.strip().strip('"').strip("'")
      os.environ[key] = value
if not os.getenv("ENABLE_RELOCATABLE_RESOURCES"):
  print("\n\033[31m(ERROR)\033[0m: No relocatable GIMP build found. You can build GIMP with '-Drelocatable-bundle=yes' to make a build suitable for bundle creation.")
  sys.exit(1)


# Bundle deps and GIMP files
GIMP_SOURCE = Path(os.getenv("MESON_SOURCE_ROOT")).as_posix()

## System prefix: it is MSYSTEM_PREFIX
with open("meson-logs/meson-log.txt") as f:
  meson_log = f.read()
match = re.search(r"Main binary: (.*)\\bin\\python\.exe", meson_log)
MSYSTEM_PREFIX = match.group(1).replace("\\", "/")
## GIMP prefix: as set at meson configure time
GIMP_PREFIX = Path(os.getenv("MESON_INSTALL_DESTDIR_PREFIX")).as_posix()

## Bundle dir: we make a "perfect" bundle separated from GIMP_PREFIX
#NOTE: The bundling script need to set $MSYSTEM_PREFIX to our dist scripts
#fallback code be able to identify what arch they are distributing
GIMP_DISTRIB = Path(GIMP_SOURCE) / f"gimp-{Path(MSYSTEM_PREFIX).name}"

def bundle(src_root, pattern):
  ## Search for targets in search path
  src_root = Path(src_root)
  paths_to_bundle = list(src_root.glob(pattern))
  if not paths_to_bundle:
    print(f"\033[31m(ERROR)\033[0m: not found {src_root}/{pattern}")
    sys.exit(1)
  for src_path in paths_to_bundle:
    ## Copy found targets to bundle path
    dest_path = GIMP_DISTRIB / src_path.relative_to(src_root)
    dest_path.parent.mkdir(parents=True, exist_ok=True)
    print(f"Bundling {src_path} to {dest_path.parent}")
    if src_path.is_dir():
      shutil.copytree(src_path, dest_path, dirs_exist_ok=True)
    else:
      shutil.copy2(src_path, dest_path)
      ## Process .typelib dependencies
      if str(src_path).endswith(".typelib"):
        def process_typelib(path, typelib_list=None):
          if typelib_list is None:
            typelib_list = set()
          cmd = ['g-ir-inspect', '--print-typelibs', os.path.basename(path).split('-')[0]]
          result = subprocess.run(cmd, capture_output=True, text=True)
          for line in result.stdout.splitlines():
            typelib = line.replace("typelib: ", "").strip()
            if typelib and typelib not in typelib_list:
              typelib_list.add(typelib)
              typelib_path = Path(f"{MSYSTEM_PREFIX}/lib/girepository-1.0/{typelib}.typelib")
              if typelib_path.exists():
                shutil.copy2(typelib_path, dest_path.parent)
                process_typelib(typelib, typelib_list)
        process_typelib(src_path)

def clean(base_path, pattern):
  base_path = Path(base_path)
  first_found = False
  for parent_path in base_path.glob(os.path.dirname(pattern)):
    for path in parent_path.rglob(os.path.basename(pattern)):
      if path.exists():
        if not first_found:
           print(f"Cleaning {base_path}/{pattern}")
           first_found = True
        if path.is_dir():
          shutil.rmtree(path)
        else:
          path.unlink()


## PREPARE BUNDLE
GIMP_DISTRIB.mkdir(parents=True, exist_ok=True)
### Prevent Git going crazy
(GIMP_DISTRIB / ".gitignore").write_text("*\n")
### Add a wrapper at tree root, less messy than having to look for the
### binary inside bin/, in the middle of all the DLLs.
(GIMP_DISTRIB / "gimp.cmd").write_text(f"powershell bin\\gimp-{os.getenv('GIMP_MUTEX_VERSION')}.exe\n")


## BUNDLE BASE (BARE MINIMUM TO RUN GTK APPS).
### Needed to not pollute output. See: https://gitlab.gnome.org/GNOME/gimp/-/issues/8877
bundle(MSYSTEM_PREFIX, "bin/gdbus.exe")
### Needed for hyperlink support etc... See: https://gitlab.gnome.org/GNOME/gimp/-/issues/12288
####...when running from `gimp*.exe` on console
bundle(MSYSTEM_PREFIX, "bin/gspawn*-console.exe")
####...when running from `gimp*.exe` from shortcut
bundle(MSYSTEM_PREFIX, "bin/gspawn*-helper.exe")
### Needed for file dialogs (only .compiled file is needed on MS Windows)
bundle(MSYSTEM_PREFIX, "share/glib-*/schemas/gschemas.compiled")
### Needed to open remote files
bundle(MSYSTEM_PREFIX, "lib/gio")
### Needed to not crash UI. See: https://gitlab.gnome.org/GNOME/gimp/-/issues/6165
bundle(MSYSTEM_PREFIX, "share/icons/Adwaita")
### Needed by GTK to use icon themes. See: https://gitlab.gnome.org/GNOME/gimp/-/issues/5080
bundle(GIMP_PREFIX, "share/icons/hicolor")
### Needed to loading icons in GUI
bundle(MSYSTEM_PREFIX, "lib/gdk-pixbuf-*/*/loaders/libpixbufloader-png.dll")
bundle(MSYSTEM_PREFIX, "lib/gdk-pixbuf-*/*/loaders/pixbufloader_svg.dll")
bundle(MSYSTEM_PREFIX, "lib/gdk-pixbuf-*/*/loaders.cache")


## CORE FEATURES.
bundle(GIMP_PREFIX, "bin/libbabl*.dll")
bundle(GIMP_PREFIX, "lib/babl-*")
bundle(GIMP_PREFIX, "bin/libgegl*.dll")
bundle(GIMP_PREFIX, "lib/gegl-*")
bundle(GIMP_PREFIX, "bin/libgimp*.dll")
bundle(GIMP_PREFIX, "lib/gimp")
bundle(GIMP_PREFIX, "share/gimp")
lang_array = [Path(f).stem for f in glob(str(Path(GIMP_SOURCE)/"po/*.po"))]
for lang in lang_array:
  bundle(GIMP_PREFIX, f"share/locale/{lang}/LC_MESSAGES/*.mo")
  # Needed for eventually used widgets, GTK inspector etc
  if glob(f"{MSYSTEM_PREFIX}/share/locale/{lang}/LC_MESSAGES/gtk*.mo"):
    bundle(MSYSTEM_PREFIX, f"share/locale/{lang}/LC_MESSAGES/gtk*.mo")
  # For language list in text tool options
  if glob(f"{MSYSTEM_PREFIX}/share/locale/{lang}/LC_MESSAGES/iso_639_3.mo"):
    bundle(MSYSTEM_PREFIX, f"share/locale/{lang}/LC_MESSAGES/iso_639_3.mo")
bundle(GIMP_PREFIX, "etc/gimp")


## OTHER FEATURES AND PLUG-INS.
### Support for legacy Win clipboard images: https://gitlab.gnome.org/GNOME/gimp/-/issues/4802
bundle(MSYSTEM_PREFIX, "lib/gdk-pixbuf-*/*/loaders/libpixbufloader-bmp.dll")
### Support for non .PAT patterns: https://gitlab.gnome.org/GNOME/gimp/-/issues/12351
bundle(MSYSTEM_PREFIX, "lib/gdk-pixbuf-*/*/loaders/libpixbufloader-jpeg.dll")
bundle(MSYSTEM_PREFIX, "lib/gdk-pixbuf-*/*/loaders/libpixbufloader-gif.dll")
bundle(MSYSTEM_PREFIX, "lib/gdk-pixbuf-*/*/loaders/libpixbufloader-tiff.dll")
### mypaint brushes
bundle(MSYSTEM_PREFIX, "share/mypaint-data/2.0")
### Needed for fontconfig
bundle(MSYSTEM_PREFIX, "etc/fonts")
#### Avoid other programs breaking the cache. See: https://gitlab.gnome.org/GNOME/gimp/-/issues/1366
fonts_conf = GIMP_DISTRIB / "etc/fonts/fonts.conf"
text = fonts_conf.read_text()
new_text = text.replace(
  "LOCAL_APPDATA_FONTCONFIG_CACHE",
  f"~/AppData/Local/GIMP/{os.getenv('GIMP_APP_VERSION')}/fontconfig/cache"
)
fonts_conf.write_text(new_text)
### Needed for 'th' word breaking in Text tool etc
bundle(MSYSTEM_PREFIX, "share/libthai")
### Needed for full CJK and Cyrillic support in file-pdf
bundle(MSYSTEM_PREFIX, "share/poppler")
### Needed for file-wmf work
bundle(MSYSTEM_PREFIX, "share/libwmf")
### Needed for 'Show image graph'
### if show_debug_menu is true in app/main.c or --show-debug-menu CLI option is set
#### See: https://gitlab.gnome.org/GNOME/gimp/-/issues/6045
bundle(MSYSTEM_PREFIX, "bin/dot.exe")
#### See: https://gitlab.gnome.org/GNOME/gimp/-/issues/12119
bundle(MSYSTEM_PREFIX, "bin/libgvplugin_dot*.dll")
bundle(MSYSTEM_PREFIX, "bin/libgvplugin_pango*.dll")
bundle(MSYSTEM_PREFIX, "bin/config8")
### Binaries for GObject Introspection support. See: https://gitlab.gnome.org/GNOME/gimp/-/issues/13170
bundle(GIMP_PREFIX, "lib/girepository-*/*.typelib")
bundle(MSYSTEM_PREFIX, "bin/libgirepository-*.dll")
#### Python support
#####python.exe is needed for plug-ins error output if `gimp*.exe` is run from console
bundle(MSYSTEM_PREFIX, "bin/python.exe")
#####pythonw.exe is needed to run plug-ins silently if `gimp*.exe` is run from shortcut
bundle(MSYSTEM_PREFIX, "bin/pythonw.exe")
bundle(MSYSTEM_PREFIX, "lib/python*")
clean(GIMP_DISTRIB, "lib/python*/*.pyc")
#####avoid lib_bundle.py bundling build-time libLLVM*.dll which is giant
clean(GIMP_DISTRIB, "lib/python*/site-packages/lldb*")
#####Needed for internet connection on python. See: https://gitlab.gnome.org/GNOME/gimp/-/issues/14722
bundle(MSYSTEM_PREFIX, "etc/ssl/cert.pem")
#### FIXME: luajit crashes at startup: See: https://gitlab.gnome.org/GNOME/gimp/-/issues/11597
#bundle(MSYSTEM_PREFIX, "bin/luajit.exe")
#bundle(MSYSTEM_PREFIX, "lib/lua")
#bundle(MSYSTEM_PREFIX, "share/lua")


## MAIN EXECUTABLES AND DEPENDENCIES
### Minimal (and some additional) executables for the 'bin' folder
bundle(GIMP_PREFIX, "bin/gimp*.exe")
### Bundled just to promote GEGL. See: https://gitlab.gnome.org/GNOME/gimp/-/issues/10580
bundle(GIMP_PREFIX, "bin/gegl.exe")
### Deps (DLLs) of the binaries in 'bin' and 'lib' dirs
### We save the list of already copied DLLs to keep a state between 2_bundle-gimp-uni_dep runs.
done_dll = Path(f"{os.getenv('MESON_BUILD_ROOT')}/done-dll.list")
done_dll.unlink(missing_ok=True)
for dir in ["bin", "lib"]:
  search_dir = GIMP_DISTRIB / dir
  print(f"Searching for dependencies of {search_dir} in {GIMP_PREFIX} and {MSYSTEM_PREFIX}")
  for ext in ("*.dll", "*.exe"):
    for dep in search_dir.rglob(ext):
      subprocess.run([
        sys.executable, f"{GIMP_SOURCE}/tools/lib_bundle.py",
        str(dep), f"{GIMP_PREFIX}/", f"{MSYSTEM_PREFIX}/",
        str(GIMP_DISTRIB), "--output-dll-list", done_dll.as_posix()
      ], check=True)


## .PDB/CODEVIEW DEBUG SYMBOLS (from babl, gegl and GIMP binaries)
bundle(GIMP_PREFIX, "bin/*.pdb")
### Remove .pdb without corresponding binaries (depends on what was choosen to be bundled above)
files = os.listdir(GIMP_DISTRIB / "bin")
binaries = {os.path.splitext(file)[0] for file in files if file.endswith(('.exe', '.dll'))}
for file in files:
  if file.endswith('.pdb'):
    if os.path.splitext(file)[0] not in binaries:
      print(f"(INFO): removing orphan {file}")
      os.remove(os.path.join(GIMP_DISTRIB / "bin", file))
### Remove ancient COFF symbol table (not used for debugging) from MSYS2 binaries
for bin_path in GIMP_DISTRIB.rglob("*"):
  if bin_path.suffix.lower() in (".dll", ".exe", ".pyd"):
    with open(bin_path, "rb") as f:
      if b"RSDS" not in f.read():
        print(f"(INFO): stripping COFF symbols from {bin_path.name}")
        try:
          subprocess.run(["strip", str(bin_path)], check=True)
        except:
          continue


## DEVELOPMENT FILES (to build GEGL filters and GIMP plug-ins).
clean(GIMP_DISTRIB, "lib/*.a")
bundle(GIMP_PREFIX, "include/gimp-*")
bundle(GIMP_PREFIX, "include/babl-*")
bundle(GIMP_PREFIX, "include/gegl-*")
bundle(GIMP_PREFIX, "lib/libgimp*.a")
bundle(GIMP_PREFIX, "lib/libbabl*.a")
bundle(GIMP_PREFIX, "lib/libgegl*.a")
bundle(GIMP_PREFIX, "lib/pkgconfig/gimp*")
bundle(GIMP_PREFIX, "lib/pkgconfig/babl*")
bundle(GIMP_PREFIX, "lib/pkgconfig/gegl*")



#!/usr/bin/env python3

import inspect
import sys

gimp_test_filename = ''

def gimp_assert(subtest_name, test):
  '''
  Please call me like this, for instance, if I were testing if gimp_image_new()
  succeeded:
    gimp_assert("gimp_image_new()", image is not None)
  '''
  if not test:
    frames = inspect.getouterframes(inspect.currentframe())
    sys.stderr.write("\n**** START FAILED SUBTEST *****\n")
    sys.stderr.write("ERROR: {} - line {}: {}\n".format(gimp_test_filename,
                                                        frames[1].lineno,
                                                        subtest_name))
    sys.stderr.write("***** END FAILED SUBTEST ******\n\n")
  assert test

def gimp_c_assert(c_filename, error_msg, test):
  '''
  This is called by the platform only, and print out the GError message from the
  C test plug-in.
  '''
  if not test:
    sys.stderr.write("\n**** START FAILED SUBTEST *****\n")
    sys.stderr.write("ERROR: {}: {}\n".format(c_filename, error_msg))
    sys.stderr.write("***** END FAILED SUBTEST ******\n\n")
  assert test



#!/usr/bin/env python3

import sys, os
from pathlib import Path
import argparse

import json

import shutil


class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)

class MesonStatus(metaclass = Singleton):
    def __init__(self):
        self.get_build_dir()
        self.get_meson_info()
        self.get_meson_installed()

    def get_build_dir(self):
        self.buildroot = None

        # Set up by meson.
        cwd = Path(os.environ['MESON_BUILD_ROOT'])

        if (cwd / 'meson-info').exists():
            self.buildroot = cwd

        if self.buildroot is None:
            print('Error: build directory was not detected. Are you in it ?')

    def get_meson_info(self):
        with open(self.buildroot / 'meson-info' / 'meson-info.json') as file:
            self.meson_info = json.load(file)
        self.sourceroot = Path(self.meson_info['directories']['source'])

    def get_meson_installed(self):
        info = self.meson_info['directories']['info']
        inst = self.meson_info['introspection']['information']['installed']['file']

        with open(Path(info) / inst) as file:
            self.installed_files = json.load(file)


def get_files_of_part(part_name):
    files_of_part = {}
    sourcedir = str(MesonStatus().sourceroot / part_name)
    builddir  = str(MesonStatus().buildroot  / part_name)
    for file, target in MesonStatus().installed_files.items():
        if file.startswith(sourcedir + '/') or file.startswith(builddir + '/'):
            files_of_part[file] = target

    return files_of_part

def install_files(files, verbose):
    warnings = []
    for file in sorted(files.keys()):
        target = files[file]
        if verbose: print(file + '  →  ' + target, end='\n')

        if os.path.isdir(file):
            copytree(file, target)
        if os.path.isfile(file):
            try:
                shutil.copy2(file, target)
            except Exception as e:
                warnings += [(file, e)]
    if len(warnings) > 0:
        sys.stderr.write("\n*** WARNING: *** Some file installation failed:\n")
        for (file, e) in warnings:
            sys.stderr.write("- {}: {}\n".format(file, e))

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('subdirs', nargs='+')
    parser.add_argument('--verbose', '-v', action='store_true')
    args = parser.parse_args()

    verbose = args.verbose

    for subdir in args.subdirs:
        files = get_files_of_part(subdir)
        if len(files) == 0:
            print('Error: subdir %s does not contain any installable file' % subdir)
        install_files(files, verbose)



#!/usr/bin/python3

# Equivalent to:
# configure_file(input: src,
#                output: name / src,
#                copy: true,
#                install_dir: gimpplugindir / 'plug-ins' / name,
#                install_mode: 'rwxr-xr-x')
# Except that configure_file() does not accept output in a subdirectory. So we
# use this wrapper for now.
# See: https://github.com/mesonbuild/meson/issues/2320
import os
import shutil
import stat
import sys

src_file = sys.argv[1]
dir_name = sys.argv[2]
dummy_path = None
if len(sys.argv) > 3:
  dummy_path = sys.argv[3]

os.makedirs(dir_name, exist_ok=True)

file_name = os.path.basename(src_file)
dst_file  = os.path.join(dir_name, file_name)
shutil.copyfile(src_file, dst_file)
os.chmod(dst_file, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)

if dummy_path is not None:
  # Just touch the dummy file.
  open(dummy_path, mode='w').close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   GIMP - The GNU Image Manipulation Program
#   Copyright (C) 1995 Spencer Kimball and Peter Mattis
#
#   gimpexporttests.py
#   Copyright (C) 2021-2024 Jacob Boerema
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module where we add export tests for GIMP's file export plug-ins."""

import sys

import gi
gi.require_version('Gimp', '3.0')
from gi.repository import Gimp
from gi.repository import Gio

from gimptestframework import GimpExportTest


class BmpExportTests(GimpExportTest):
    def __init__(self, group_name, log):
        super().__init__(group_name, log)

    def load_bmp(self, bmp_name):
        pdb_proc   = Gimp.get_pdb().lookup_procedure(self.file_import)
        pdb_config = pdb_proc.create_config()
        pdb_config.set_property('run-mode', Gimp.RunMode.NONINTERACTIVE)
        pdb_config.set_property('file', Gio.File.new_for_path(bmp_name))
        result = pdb_proc.run(pdb_config)
        status = result.index(0)
        img = result.index(1)

        if (status == Gimp.PDBStatusType.SUCCESS and img is not None and img.is_valid()):
            return img
        else:
            self.log.error(f"Loading image {bmp_name} failed!")
            return None

    def export_bmp(self, img, bmp_exp_name, rle, space, rgb):
        drw = img.get_layers()
        export_file_path = self.output_path + bmp_exp_name

        pdb_proc   = Gimp.get_pdb().lookup_procedure(self.file_export)
        pdb_config = pdb_proc.create_config()
        pdb_config.set_property('run-mode', Gimp.RunMode.NONINTERACTIVE)
        pdb_config.set_property('image', img)
        pdb_config.set_property('num-drawables', 1)
        pdb_config.set_property('drawables', Gimp.ObjectArray.new(Gimp.Drawable, drw, False))
        pdb_config.set_property('file', Gio.File.new_for_path(export_file_path))
        pdb_config.set_property('use-rle', rle)
        pdb_config.set_property('write-color-space', space)
        pdb_config.set_property('rgb-format', rgb)

        result = pdb_proc.run(pdb_config)
        status = result.index(0)
        if status == Gimp.PDBStatusType.SUCCESS:
            self.log.info(f"Exporting image {export_file_path} succeeded.")
        else:
            self.log.error(f"Exporting image {export_file_path} failed!")

    def bmp_setup(self, image_name, test_description):
        self.log.consoleinfo(test_description)
        tmp = self.input_path + image_name
        img = self.load_bmp(tmp)
        return img

    def bmp_teardown(self, image):
        if image is not None:
            image.delete()

    def bmp_1bpp_indexed(self, image_name, test_description):
        img = self.bmp_setup(image_name, test_description)
        if img is not None:
            self.export_bmp(img, "1bpp-norle-nospace.bmp", 0, 0, 0)
            self.export_bmp(img, "1bpp-norle-withspace.bmp", 0, 1, 0)
            # rle with 1-bit is not valid, test to see if it
            # is correctly ignored...
            self.export_bmp(img, "1bpp-rle-nospace.bmp", 1, 0, 0)
            self.export_bmp(img, "1bpp-rle-withspace.bmp", 1, 1, 0)
        self.bmp_teardown(img)

    def bmp_4bpp_indexed(self, image_name, test_description):
        img = self.bmp_setup(image_name, test_description)
        if img is not None:
            self.export_bmp(img, "4bpp-norle-nospace.bmp", 0, 0, 0)
            self.export_bmp(img, "4bpp-norle-withspace.bmp", 0, 1, 0)
            self.export_bmp(img, "4bpp-rle-nospace.bmp", 1, 0, 0)
            self.export_bmp(img, "4bpp-rle-withspace.bmp", 1, 1, 0)
        self.bmp_teardown(img)

    def bmp_8bpp_indexed(self, image_name, test_description):
        img = self.bmp_setup(image_name, test_description)
        if img is not None:
            self.export_bmp(img, "8bpp-norle-nospace.bmp", 0, 0, 0)
            self.export_bmp(img, "8bpp-norle-withspace.bmp", 0, 1, 0)
            self.export_bmp(img, "8bpp-rle-nospace.bmp", 1, 0, 0)
            self.export_bmp(img, "8bpp-rle-withspace.bmp", 1, 1, 0)
        self.bmp_teardown(img)

    def bmp_rgba(self, image_name, test_description):
        img = self.bmp_setup(image_name, test_description)
        if img is not None:
            # first test correctly ignoring rle
            self.export_bmp(img, "rgb-with-rle-nospace.bmp", 1, 0, 3)
            # All rgb modes without color space
            self.export_bmp(img, "rgb-nrle-nospace-type0.bmp", 0, 0, 0)
            self.export_bmp(img, "rgb-nrle-nospace-type1.bmp", 0, 0, 1)
            self.export_bmp(img, "rgb-nrle-nospace-type2.bmp", 0, 0, 2)
            self.export_bmp(img, "rgb-nrle-nospace-type3.bmp", 0, 0, 3)
            self.export_bmp(img, "rgb-nrle-nospace-type4.bmp", 0, 0, 4)
            self.export_bmp(img, "rgb-nrle-nospace-type5.bmp", 0, 0, 5)
            # All rgb modes with color space
            self.export_bmp(img, "rgb-nrle-withspace-type0.bmp", 0, 1, 0)
            self.export_bmp(img, "rgb-nrle-withspace-type1.bmp", 0, 1, 1)
            self.export_bmp(img, "rgb-nrle-withspace-type2.bmp", 0, 1, 2)
            self.export_bmp(img, "rgb-nrle-withspace-type3.bmp", 0, 1, 3)
            self.export_bmp(img, "rgb-nrle-withspace-type4.bmp", 0, 1, 4)
            self.export_bmp(img, "rgb-nrle-withspace-type5.bmp", 0, 1, 5)
        self.bmp_teardown(img)

    def run_test(self, group_config):
        self.bmp_1bpp_indexed('test1-exp1.bmp',   "    1. Using 1bpp indexed BMP...")
        self.bmp_4bpp_indexed('pal4rle-exp1.bmp', "    2. Using 4bpp indexed BMP...")
        self.bmp_8bpp_indexed('pal8v5-exp1.bmp',  "    3. Using 8bpp indexed BMP...")
        self.bmp_rgba('rgb32-with-alpha-1.bmp',   "    4. Using rgb(a) BMP...")




"""
Python scripts to test libgimp API.
Specifically, test GimpResource class and subclasses.

Test methods common to most resource classes, and class methods.
Methods specific to each class are tested elsewhere.

Usage:
=====

This is not a GIMP plugin.
Just copy and paste into the GIMP Python Console.
Expect no exceptions.
You can quickly look for red text.
There are no messages from  the test program,
only exceptions from assert to indicate a failure.

Requires PyGObject binding to libgimp,
which the GIMP Python Console does automatically.

Trash:
=====

This test program may leave trash, especially for failed runs.
Trash means resources persisted by GIMP.
!!! Do not run this in a production environment (using GIMP to create graphics.)
You can run it in a GIMP development environment,
but you may need to clean up trash manually.
Deleting the .config/GIMP directory will clean up the trash.

Setup in the GIMP app:
=====================

Choose current resources to match what this test uses:
e.g. brush '2. Hardness 050'
Use the standard default initial resources for a clean build,
by deleting .config/GIMP.

Ensure test resources from prior runs are deleted
otherwise they interfere with testing i.e. name clash.
E.g. delete brush  '2. Hardness 050 copy #1'
"""


"""
Test return value is-a resource.
Get resource from context
"""
brush    = Gimp.context_get_brush()
font     = Gimp.context_get_font()
gradient = Gimp.context_get_gradient()
palette  = Gimp.context_get_palette()
pattern  = Gimp.context_get_pattern()

assert isinstance(brush,    Gimp.Brush)
assert isinstance(font,     Gimp.Font)
assert isinstance(gradient, Gimp.Gradient)
assert isinstance(palette,  Gimp.Palette)
assert isinstance(pattern,  Gimp.Pattern)

"""
Test is_valid method
Expect True
"""
assert brush.is_valid()
assert font.is_valid()
assert gradient.is_valid()
assert palette.is_valid()
assert pattern.is_valid()

"""
Test get_id method
Expect string
"""
print( brush.get_id() )
print( font.get_id() )
print( gradient.get_id() )
print( palette.get_id() )
print( pattern.get_id() )

assert brush.get_id()    == '2. Hardness 050'
assert font.get_id()     == 'Sans-serif'
assert gradient.get_id() == 'FG to BG (RGB)'
assert palette.get_id()  == 'Color History'
assert pattern.get_id()  == 'Pine'

"""
Test resource as an arg
Pass the resource back to the context
Expect no exceptions
"""
assert Gimp.context_set_brush(brush)        == True
assert Gimp.context_set_font(font)          == True
assert Gimp.context_set_gradient(gradient)  == True
assert Gimp.context_set_palette(palette)    == True
assert Gimp.context_set_pattern(pattern)    == True

"""
Test new() methods

Not Font, Pattern.
Pass desired name.
"""
brush_new = Gimp.Brush.new("New Brush")
# expect a valid brush object, not a name
assert brush_new.is_valid()
# expect id is the desired one (assuming no clash)
assert brush_new.get_id()=="New Brush"

gradient_new = Gimp.Gradient.new("New Gradient")
assert gradient_new.is_valid()
assert gradient_new.get_id()=="New Gradient"

palette_new = Gimp.Palette.new("New Palette")
assert palette_new.is_valid()
assert palette_new.get_id()=="New Palette"

"""
Test delete methods.
Delete the new resources we just made
After deletion, reference is invalid
"""
assert brush_new.delete() == True
assert brush_new.is_valid()==False

assert gradient_new.delete() == True
assert gradient_new.is_valid()==False

assert palette_new.delete() == True
assert palette_new.is_valid()==False

"""
Test copy methods.

Copy the original resources from context.
Assert that GIMP adds " copy"
"""
brush_copy = brush.duplicate()
assert brush_copy.is_valid()
assert brush_copy.get_id()=='2. Hardness 050 copy'

gradient_copy = gradient.duplicate()
assert gradient_copy.is_valid()
assert gradient_copy.get_id()=='FG to BG (RGB) copy'

palette_copy = palette.duplicate()
assert palette_copy.is_valid()
assert palette_copy.get_id()== 'Color History copy'


"""
Test rename methods.

Renaming returns a reference that must be kept. !!!

Rename existing copy to a desired name.
!!! assign to a new named var
"""
brush_renamed = brush_copy.rename("Renamed Brush")
assert brush_renamed.is_valid()
assert brush_renamed.get_id()=="Renamed Brush"
# assert renaming invalidates any old reference,
assert brush_copy.is_valid()==False

gradient_renamed = gradient_copy.rename("Renamed Gradient")
assert gradient_renamed.is_valid()
assert gradient_renamed.get_id()=="Renamed Gradient"
# assert renaming invalidates any old reference,
assert gradient_copy.is_valid()==False

palette_renamed = palette_copy.rename("Renamed Palette")
assert palette_renamed.is_valid()
assert palette_renamed.get_id()=="Renamed Palette"
# assert renaming invalidates any old reference,
assert palette_copy.is_valid()==False


"""
Test renaming when name is already in use.
If "Renamed Brush" is already in use, then the name will be
"Renamed Brush copy #1"
that is, renaming added suffix "#1"

This is due to core behavior, not libgimp resource class per se,
so we don't test it here.
"""

"""
Test delete methods
Delete the renamed copies we just made
After deletion, reference is invalid
"""
brush_renamed.delete()
assert brush_renamed.is_valid()==False

gradient_renamed.delete()
assert gradient_renamed.is_valid()==False

palette_renamed.delete()
assert palette_renamed.is_valid()==False


'''
Test class method

!!! Expect no error dialog in GIMP.
The app code generated from foo.pdb should clear the GError.
'''
assert Gimp.Brush.id_is_valid   ("invalid name")==False
assert Gimp.Font.id_is_valid    ("invalid name")==False
assert Gimp.Gradient.id_is_valid("invalid name")==False
assert Gimp.Palette.id_is_valid ("invalid name")==False
assert Gimp.Pattern.id_is_valid ("invalid name")==False

'''
Test constructor
Not something author's should do, but test it anyway
'''
brush2    = Gimp.Brush()
font2     = Gimp.Font()
gradient2 = Gimp.Gradient()
palette2  = Gimp.Palette()
pattern2  = Gimp.Pattern()
assert isinstance(brush2,    Gimp.Brush)
assert isinstance(font2,     Gimp.Font)
assert isinstance(gradient2, Gimp.Gradient)
assert isinstance(palette2,  Gimp.Palette)
assert isinstance(pattern2,  Gimp.Pattern)

# a created instance is invalid until it's ID is set
assert brush2.is_valid()==False
assert font2.is_valid()==False
assert gradient2.is_valid()==False
assert palette2.is_valid()==False
assert pattern2.is_valid()==False

"""
test set_property methods

Use a name that exists.  Creating de novo a reference to an existing brush.
Reference is valid since it's data exists in app core
"""

brush2.set_property("id", "2. Star")
assert brush2.get_id()=="2. Star"
assert brush2.is_valid()

font2.set_property("id", "Sans-serif")
assert font2.get_id()  =="Sans-serif"
assert font2.is_valid()

gradient2.set_property("id", 'FG to BG (RGB)')
assert gradient2.get_id()  =='FG to BG (RGB)'
assert gradient2.is_valid()

palette2.set_property("id", 'Color History')
assert palette2.get_id()  =='Color History'
assert palette2.is_valid()

pattern2.set_property("id", 'Pine')
assert pattern2.get_id()  =='Pine'
assert pattern2.is_valid()
